#include <sys/times.h>
#include <iostream>
#include <fstream>
#include <string>
#include <math.h>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_cdf.h>

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/boyer_myrvold_planar_test.hpp>

using namespace boost;

typedef boost::adjacency_list<boost::vecS, boost::vecS, boost::undirectedS, boost::property<vertex_index_t, int> > Graph;
typedef graph_traits<Graph>::edge_iterator EdgeIt;
typedef graph_traits<Graph>::out_edge_iterator OutEdgeIt;
typedef graph_traits<Graph>::vertex_iterator VertexIt;
typedef graph_traits<Graph>::edge_descriptor EdgeD;

template<typename Graph, typename EdgeIt, typename VertexIt>
void g2Dot(Graph g, std::string filename)
{
  std::ofstream fout(filename.c_str());
  fout << "graph G {" << std::endl;
  fout << "graph[size=\"100,100\"];" << std::endl;

  // var nodes 
  fout << "node[shape=circle,width=0.25,height=0.25,fixedsize=true];" << std::endl;
	VertexIt vi, ve;
	for (tie(vi, ve) = vertices(g); vi != ve; vi++) {
    fout << "\t" << *vi;
    fout << " [style=\"setlinewidth(1)\",label=" << *vi << "];" << std::endl;
  }

	// edges 
  EdgeIt ei, ei_end;
  for (tie(ei, ei_end) = edges(g); ei != ei_end; ei++) {
		fout << "\t" << source(*ei, g) << " -- " << target(*ei, g);
		fout << "[style=\"setlinewidth(1)\",fontsize=15,label=\"";
		fout << "\"];" << std::endl;
  }
  fout << "}" << std::endl;
  fout.close();
}

void writeFg(const Graph &g, std::string filename, gsl_rng *r, double sigma, double theta)
{
	std::ofstream fout( filename.c_str() );
	fout << num_edges(g) + num_vertices(g) << std::endl << std::endl;// + num_vertices(g)

	// pairwise potentials
	VertexIt vb, ve;
	for (tie(vb, ve) = vertices(g); vb != ve; vb++) {
		OutEdgeIt eb, ee;
		for (tie(eb, ee) = out_edges(*vb, g); eb != ee; eb++) {

			if (*vb < target(*eb, g)) {

				fout << 2 << std::endl;
				fout << *vb << " " << target(*eb, g) << std::endl;
				fout << "2 2" << std::endl<< 4 << std::endl;
		
				double w = gsl_ran_gaussian(r, sigma);
		
				fout << "0\t" << exp(w) << std::endl;
				fout << "1\t" << exp(-w) << std::endl;
				fout << "2\t" << exp(-w) << std::endl;
				fout << "3\t" << exp(w) << std::endl << std::endl;
			}
		}

		// single variable potential
		fout << 1 << std::endl;
		fout << *vb << std::endl;
		fout << "2" << std::endl<< 2 << std::endl;
		
		double w = gsl_ran_gaussian(r, theta);

		fout << "0\t" << exp(w) << std::endl;
		fout << "1\t" << exp(-w) << std::endl << std::endl;
	}
	fout.close();
}

int main( int argc, char *argv[] ) {
	int s, n;
	double sigma, theta;

	if (argc != 6) {
		std::cerr << "./create_phex filename sides nesting theta sigma" << std::endl;
		exit(-1);
	}

	std::string sfilename(argv[1]);
	s = atoi(argv[2]);
	n = atoi(argv[3]);
	theta = atof(argv[4]);
	sigma = atof(argv[5]);

	std::cout << "filename         " << sfilename << std::endl;
	std::cout << "sides            " << s << std::endl;
	std::cout << "nesting          " << n << std::endl;
	std::cout << "theta            " << theta << std::endl;
	std::cout << "sigma            " << sigma << std::endl;
	
	// Init pseudorandom generator
  gsl_rng_env_setup();
  gsl_rng *r = gsl_rng_alloc(gsl_rng_mrg);
 
	// create nested polygons
	Graph g;
	for (int i=0; i<n; i++) {
		for (int j=0; j<s; j++) {
			add_vertex(g);
			if (i) add_edge((i-1)*s + j, i*s + j, g);
			if (j){
				add_edge(i*s+j-1, i*s+j, g);
				std::cout << i*s+j-1 << " ~ " << i*s+j << std::endl;
			}
		}
		add_edge(i*s, i*s+s-1, g);
		std::cout << i*s << " ~ " << i*s+s-1 << std::endl;


	}
//	int N = num_vertices(g);

	if (boyer_myrvold_planarity_test(g)) {
	
		// write .fg graph
		char buf[100];
		sprintf(buf, "%s_0.fg", sfilename.c_str());
		std::cout << "creating " << buf << std::endl;
		writeFg(g, std::string(buf), r, sigma, theta);
	
		// write .dot graph
		sprintf(buf, "%s_0.dot", sfilename.c_str());
		std::cout << "creating " << buf << std::endl;
		g2Dot<Graph, EdgeIt, VertexIt>(g, std::string(buf) );
	}
	else { std::cout << "Initial graph is not planar..." << std::endl; exit(-1); }

  return 0;
}
