#ifndef __FACTOR_H__
#define __FACTOR_H__

/*
 *		factor ();
 *		factor (const myset& set);
 *		factor (const myset& set, double z);
 *		factor (const myset& set, const double *p);
 *		factor (El a, El b, ..., const double *p);
 *		factor (El a, El b, ..., double p00, double p01, ...);
 *		logd operator [] (unsigned long s) const;
 *		logd& operator [] (unsigned long s);
 *		factor& operator &= (const myset& set);
 *		factor& operator |= (const factor& f);
 *		factor& operator /= (const myset& set);
 *		factor& operator <<= (const factor& f);
 *		factor operator & (const myset& set) const;
 *		factor operator | (const factor& f) const;
 *		factor operator / (const myset& set) const;
 */

#include "logd.h"
#include "myset.h"
#include "index.h"
#include <vector>

/* These macro's are used to define a lot of member functions for simple
 * creation of factors.  Example: factor<El>(a,b, 0.1,0.3,0.2,0.4) creates
 * a factor for nodes a and b with probability table [0.1,0.3,0.2,0.4]
 */

#define XARG1(El) const El& a
#define XCALL1 a
#define XARG2(El) XARG1(El), const El& b
#define XCALL2 a, b
#define XARG3(El) XARG2(El), const El& c
#define XCALL3 a, b, c
#define XARG4(El) XARG3(El), const El& d
#define XCALL4 a, b, c, d
#define XARG5(El) XARG4(El), const El& e
#define XCALL5 a, b, c, d, e
#define XARG6(El) XARG5(El), const El& f
#define XCALL6 a, b, c, d, e, f

#define PARG1 double p0, double p1
#define PCALL1 p0, p1
#define PARG2 double p00, double p01, double p10, double p11
#define PCALL2 p00, p01, p10, p11
#define PARG3 \
		double p000, double p001, double p010, double p011, \
		double p100, double p101, double p110, double p111
#define PCALL3 p000, p001, p010, p011, p100, p101, p110, p111
#define PARG4 \
		double p0000, double p0001, double p0010, double p0011, \
		double p0100, double p0101, double p0110, double p0111, \
		double p1000, double p1001, double p1010, double p1011, \
		double p1100, double p1101, double p1110, double p1111
#define PCALL4 \
		p0000, p0001, p0010, p0011, p0100, p0101, p0110, p0111, \
		p1000, p1001, p1010, p1011, p1100, p1101, p1110, p1111

#define CONSTRUCTOR_SHORT(Class, El, n) \
		Class (XARG##n(El), const double *p) \
			{ El x[] = {XCALL##n}; _setup(n, x, p); };
#define CONSTRUCTOR_LONG(Class, El, n) \
		Class (XARG##n(El), PARG##n) \
			{ El x[] = {XCALL##n}; double p[] = {PCALL##n}; _setup(n, x, p); };

#define FACTOR_CONSTRUCTORS(Class, El) \
		CONSTRUCTOR_SHORT(Class, El, 1) CONSTRUCTOR_SHORT(Class, El, 2) \
		CONSTRUCTOR_SHORT(Class, El, 3) CONSTRUCTOR_SHORT(Class, El, 4) \
		CONSTRUCTOR_SHORT(Class, El, 5) CONSTRUCTOR_SHORT(Class, El, 6) \
		CONSTRUCTOR_LONG(Class, El, 1) CONSTRUCTOR_LONG(Class, El, 2) \
		CONSTRUCTOR_LONG(Class, El, 3) CONSTRUCTOR_LONG(Class, El, 4)

namespace std {

template<class El, class Cmp = less<El> >
class factor: public myset<El, Cmp> {
private:
	typedef myset<El, Cmp> myset;
	typedef vector<logd> _table;
	_table _p;
public:
	factor (): myset() {
		_p.resize(1ul << this->size());
	};
	factor (const myset& set): myset(set) {
		_p.resize(1ul << this->size());
	};
	factor (const myset& set, double z): myset(set) {
		_p.resize(1ul << this->size(), logd(z));
	};
	factor (const myset& set, const double *p): myset(set) {
		_p.resize(1ul << this->size());
		for(unsigned long s = _p.size(); s--; ) _p[s] = logd(p[s]);
	};

protected:
	void _setup (int n, const El *x, const double *p) {
		_p.resize(1ul << n);
		for(int i = n; i--; ) {
			if((*this) && x[i]) {
				cerr << "error: node set contains duplicates" << endl;
				exit(1);
			};
			insert(x[i]);
		};
		for(index i((myset)(*this), x, x + n); i.running(); i.next())
			_p[i.full()] = logd(p[i.sub()]);
	};

public:
	factor (int n, const El *x, const double *p) { _setup(n, x, p); };
	FACTOR_CONSTRUCTORS(factor, El)

	inline logd operator [] (unsigned long s) const { return(_p[s]); };
	inline logd& operator [] (unsigned long s) { return(_p[s]); };

	factor& operator &= (const myset& set) {
		return((*this) = (*this) & set);
	};
	factor& operator |= (const factor& f) {
		return((*this) = (*this) | f);
	};
	factor& operator /= (const myset& set) {
		return((*this) = (*this) / set);
	};
	// this function assumes that f is a subset of this
	// it computes this(a,b) = this(a,b) * f(b)
	factor& operator <<= (const factor& f) {
		for(index i((myset)*this, (myset)f); i.running(); i.next())
			(*this)[i.full()] *= f[i.sub()];
		return(*this);
	};

  // Vicen
	double Z () {
		logd Z(_p[0]);
	  for(unsigned long s = 1; s<_p.size(); s++) Z += _p[s];
		return Z;
	}

	factor& collect (const factor& f) {
		_p.clear(); _p.resize(1ul << this->size(), logd(0.0));
		for(index i((myset)f, (myset)*this); i.running(); i.next())
			(*this)[i.sub()] += f[i.full()];
		return(*this);
	};
	factor& distribute (const factor& f) {
		_table inv(_p);
		_p.clear(); _p.resize(1ul << this->size(), logd(0.0));
		for(index i((myset)f, (myset)*this); i.running(); i.next())
			(*this)[i.sub()] += f[i.full()];
		/* in normal junction tree message passing, inv[s] can only be 0 if
		 * (*this)[s] is zero as well
		 */
		for(unsigned long s = (1ul << this->size()); s--; )
			if(inv[s] != logd(0.0)) (*this)[s] /= inv[s];
		return(*this);
	};

	factor operator & (const myset& set) const {
		factor f((myset)(*this) & set, 0.0);
		for(index i(*this, f); i.running(); i.next())
			f[i.sub()] += (*this)[i.full()];
		return(f);
	};
	factor operator | (const factor& f) const {
		factor g((myset)(*this) | (myset)f);
		for(index2 i(g, *this, f); i.running(); i.next())
			g[i.full()] = (*this)[i.sub1()] * f[i.sub2()];
		return(g);
	};
	factor operator / (const myset& set) const {
		factor f((myset)(*this) / set, 0.0);
		for(index i(*this, f); i.running(); i.next())
			f[i.sub()] += (*this)[i.full()];
		return(f);
	};

	ostream& out_short (ostream& os) const {
		return(os << (myset)*this);
	};
	ostream& out_long (ostream& os) const {
		os << (myset)*this << " => [";
		for(unsigned long s = 0; s < (1ul << this->size()); ++s)
			{ if(s) os << " "; os << (*this)[s]; };
		return(os << "]");
	};
	ostream& out_matlab (ostream& os) const {
		os << "potential([";
		for(typename myset::const_iterator n = this->begin(); n != this->end(); ++n)
			{ if(n != this->begin()) os << " "; os << *n; };
		os << "], reshape([";
		for(unsigned long s = 0; s < (1ul << this->size()); ++s)
			{ if(s) os << " "; os << (*this)[s]; };
		os << "], [";
		for(int i = this->size(); i--; ) os << "2 ";
		if(this->size() == 1) os << "1 ";
		return(os << "]));");
	};
};

}; // end of namespace std

#endif

