//
// fgraph.hh
// 
// Made by vicen
// Login   <vicen@localhost.localdomain>
// 
// Started on  Wed Apr 19 20:34:02 2006 vicen
// Last update Thu Aug 31 17:56:02 2006 Vicen
//

/*! 
  \file 
  \brief
 */

#ifndef   	FGRAPH_HH_
# define   	FGRAPH_HH_

#include <iostream>                      // for std::cout
#include <fstream>                       // for std::cout
#include <utility>                       // for std::pair
#include <map>

// for random number generation
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>

#include <boost/graph/planar_canonical_ordering.hpp>
#include <boost/graph/make_biconnected_planar.hpp>
#include <boost/graph/make_maximal_planar.hpp>
#include <boost/graph/planar_face_traversal.hpp>
#include <boost/graph/boyer_myrvold_planar_test.hpp>
#include <boost/graph/chrobak_payne_drawing.hpp>
#include <boost/graph/make_connected.hpp>

#include <boost/graph/johnson_all_pairs_shortest.hpp>

#include "vprob.hh"
#include "pindex.hh"
#include "global.hh"
#include "matrix.hh"

using namespace boost;

class PfTerm
{
public:
	Real mu;
	bool sign;

	PfTerm() { mu = Real(0); sign = false; }
	PfTerm(Real _mu, bool _sign = false) { mu = _mu; sign = _sign; }
	inline PfTerm operator+ (const PfTerm &t) { return PfTerm( (*this).mu + t.mu, (*this).sign || t.sign ); }
	inline PfTerm operator- (const PfTerm &t) { return PfTerm( (*this).mu - t.mu, (*this).sign || t.sign ); }
	inline PfTerm operator- () { return PfTerm( (*this).mu, !(*this).sign ); }
	inline PfTerm operator* (const PfTerm &t) { return PfTerm( (*this).mu * t.mu, (*this).sign && t.sign ); }
	inline PfTerm operator/ (const PfTerm &t) { return PfTerm( (*this).mu / t.mu, (*this).sign && t.sign ); }
	inline bool operator== (const PfTerm &t) { return ((*this).mu == t.mu) && ((*this).sign == t.sign); }
	inline PfTerm operator= (const PfTerm &t) { if (this != &t) { mu = t.mu; sign = t.sign; }; return (*this); }
	inline Real toReal() { if (sign) return mu; else return -mu; }
  friend std::ostream& operator<< (std::ostream& os, PfTerm& pt);
};



/// The factor graph class

/**
 */

class FGraph
{
protected:

	int m_verbosity;
  gsl_rng *m_r;
  ///< random seed for gsl pseudorandom-number generation
  ///< to change: $ export GSL_RNG_SEED=1234

  Real m_max_factor_value;

	uint m_max_degree;

  int m_bp_variant;  //TBPVariant
  ///< the variant used for the bp algorithm

  MethodVariant m_method;
  ///< the variant used for computing the covariances

  uint m_nfactors;
  ///< the number of factors

  uint m_nvars;
  ///< the number of variables

  uint m_nedges;
  ///< the number of edges

  MapRes m_map_res;
  ///< the multimap for the residuals will store the residuals of each edge
  ///< sorted and is indexed from the Class PEdge using an iterator
  ///< Requirements:
  ///< - Sorted Associative Container: to get the maximum in O(1)
  ///< - Pair Associative Container: edge descriptor given the max res
  ///< - Multiple Associative Container: two edges can have same res
  ///< <b>ONLY USED IN THE MAX_RES bp_variant</b>

  std::map<int, Vertex> m_map_LR_vars;
  ///< a map for obtaining the R, covs, etc...
  ///< given the Vertex it returns the row where R, covs, etc.. are located
  ///< this map is informed at each LR iteration (see doLR method)

  std::vector<Edge> m_rnd_edges;
  ///< a random sequence of the edges
  ///< <b>ONLY USED IN THE SEQ_RND bp_variant</b>


	std::string m_filename;
  ///< filename used for the input graph and for the prefix of other output filenames
  ///< as the user entered

	bool m_output;
  ///< true if output of the beliefs required

	std::string m_filename_cov;
  ///< filename used for the covariances

  std::map<int, int> m_label2idx;
  ///< for accessing the variable index  given the label
  ///< label is arbitrary and depends on the factor graph file
  ///< Given a label, returns the index in m_vars_labels

  std::vector<Vertex> m_vars_labels;
  ///< for accessing the variable node given an index
  ///< in this vector, variable nodes are sorted by label
	///< Note that indexing is NOT per variable label

  std::map<Vertex, int> m_vars_pos;
  ///< inverse of m_vars_labels
  ///< given a variable vertex, it returns the position of m_vars_labels where it appears
  ///< (needed in the exact() method)
 
  std::vector<Vertex> m_vars_sizes;
  ///< a vector of variables sorted by their sizes

  std::vector<Vertex> m_factors;
  ///< a vector of factors

  /// creates the vector of variables sorted by size
  /** details:
   */
  void buildVarsSizes();

  /// creates the vector of variables sorted by label and the map of <variables, pos>
  /** details:
   */
  void buildVarsPos();

  /// creates the vector of factors
  /** details:
   */
  void buildFactors();

  /// adds an edge to the graph
  /** \param s the source vertex
      \param t the target vertex
      \param size the size of the BP messages
      \param pos the pos in the corresponding factor vertex
   */
  void addEdge(const Vertex &s, const Vertex &t, BGraph &_g, const uint &size, const uint &pos);

public:

  Real m_tol;
  ///< the threshold for BP to converge

  uint m_max_iter;
  ///< the maximum number of iterations until BP convergence

  Real m_maxdif;
  ///< the maximum difference between two iterations
  BGraph *g;
  ///< the boost graph

  /// default constructor
  /**
   */
  FGraph() { std::cout << "no filename, exiting ..." << std::endl; exit(-1); }

  /// constructor
  /** \param filename the filename from which the factor graph is read
      \param bp_variant the variant of the bp algorithm:
   */
  FGraph(std::string filename, TBPVariant bp_variant=BP_SEQFIX, int prec=100, int ran=0, int verbosity=0);

  void exact();

  /// Destructor
  /** destroys the boost graph object
   */
  virtual ~FGraph();

  /// returns the number of factors
  /** details:
   */
  int getNFactors() { return m_nfactors; }

  /// returns the number of variables
  /** details:
   */
  int getNVars() { return m_nvars; }

	  /// returns the number of edges
  /** details:
   */
  int getNEdges() { return num_edges(*g); }


  /// returns the vector sorted by label for accessing the nodes
  /** details:
   */
  std::vector<Vertex> getVarsLabels() { return m_vars_labels; }

  /// performs one BP n message update
  /** \param e The edge
      \return the new n message
   */
  VProb getBPn(const Edge& e, bool old);

  /// performs one BP m message update
  /** \param e The edge
      \return the new m message
   */
  VProb getBPm(const Edge& e, bool old);

  /// performs one LR N message update
  /** \param e The edge
      \param e The variable k
      \return the new N message
   */
  VProb getLRN(const Edge& e, const Vertex& k, const bool norm, int verb=0);

  /// performs one LR M message update
  /** \param e The edge
      \param e The variable k
      \return the new M message
   */
  VProb getLRM(const Edge& e, const Vertex& k, int verb=0);

  /// performs the LR algorithm
  /** \param tol The threshold (belief change) for convergence
      \param maxiter The maximum number of iterations for convergence
      \return true if convergence
   */
  bool doLR(Real tol, uint maxiter);

	/// performs the LR algorithm
  /** \param tol The threshold (belief change) for convergence
      \param maxiter The maximum number of iterations for convergence
      \return true if convergence
   */
  bool doBP(Real tol, uint maxiter, bool output=false);

  /// Returns the single node EXACT marginal
  /** \param index of the variable
   *  neither the label nor the Vertex. index is just the index of the m_vars_labels vector
   */
  VProb getExactMarginal(int index) { return (*g)[m_vars_labels[index]].p; }

  /// Returns the single node marginal of Belief Propagation
  /** \param index of the variable
   *  neither the label nor the Vertex. index is just the index of the m_vars_labels vector
   */
  VProb getBPMarginal(int index) { return (*g)[m_vars_labels[index]].bel; }

  /// Returns the single node marginal error of Belief Propagation
  /** Precondition: BP must be run before and the exact marginals should have been set using
   *  setExactMarginals or alternatively using the exact() method
   */
  Real getBPError();

  /// Returns the logZ error of Belief Propagation
  /** Precondition: BP must be run before and m_logZ_exact should have been before
   *  or alternatively the exact() method
   */
  Real getBPZError() {
    m_error_Zbp = Abs( m_logZ_bp - m_logZ_exact );
 	  return m_error_Zbp;
  }

  /// Sets the single node marginal as if it was the exact result
  /** \param index of the variable
   *  \param prob the exact marginal
   *  neither the label nor the Vertex. index is just the index of the m_vars_labels vector
   */
  void setExactMarginal(int index, const VProb &prob) { (*g)[m_vars_labels[index]].p = prob; m_exact = true; }

  /// Sets the single node marginal as if it was the exact result
  /** \param label
   *  \param prob the exact marginal
   */
  void setExactMarginalLabel(int label, const VProb &prob) { setExactMarginal(m_label2idx[label], prob); }

  void setExactLogZ(Real logZ) { m_logZ_exact = logZ; }

	/// Returns the covariances between two nodes
  /** \param label1 the variable i
      \param label2 the variable j
   */
  VProb getCov(int label_i, int label_j);

  /// Clamp / Unclamp indexing per variable index
  std::map<Vertex, VProb> clamp(int label_i, uint value);
  void unclamp(int var, std::map<Vertex, VProb> factors);

  /// Clamp / Unclamp indexing per variable label 
	std::map<Vertex, VProb> clampLabel(int label, uint val) {
		if ( (*g)[m_vars_labels[m_label2idx[label]]].label != label ) throw "Error in variable labels";
		return clamp( m_label2idx[label], val ); 
	}
  void unclampLabel(int label, std::map<Vertex, VProb> factors) {
		if ( (*g)[m_vars_labels[m_label2idx[label]]].label != label ) throw "Error in variable labels";
		unclamp( m_label2idx[label], factors );
	}
	
  /// performs the COND algorithm
  /** details:
      \return true if all BP iterations converge
   */
  bool doCOND(Real tol, uint maxiter);

  void setTBPVariant(const TBPVariant &variant) { m_bp_variant = variant; }
  TBPVariant getBPVariant() { return (TBPVariant)m_bp_variant; }

  /// sets the BP m and n messages to 1
  /** details:
   */
  void reset();

  /// outputs the covariance matrix to a file
  /** details:
   */
  void writeCov();

  /// returns the filename where the covariances have been saved
  /** details:
   */
  std::string getFilenameCov() { return m_filename_cov; }

  std::string getBPVariantString() {
    std::string ret;
    switch (m_bp_variant) {
      case BP_MAXRES: ret = "BP_MAXRES";
      case BP_SEQRND: ret = "BP_SEQRND";
      case BP_PARAL:  ret = "BP_PARAL";
      case BP_SEQFIX: ret = "BP_SEQFIX";
    }
    return ret;
  }

  /// outputs the graph to an stream
  /** details:
   */
  friend std::ostream& operator<< (std::ostream& os, FGraph& fg);

  void graph2Dot(const char *);
	void forney2Dot(const char *filename, const ForneyGraph &forney);
	void diext2Dot(const ExtDiGraph &digraph, const char *filename);
	void ext2Pigale(const ExtGraph &gext, const char *filename);
	void graph2Pigale(const char *filename);

  uint m_BP_steps;
  uint m_LR_steps;

  double m_cpu_bp;
  double m_cpu_lr;
  double m_cpu_pfaffian1;	// cpu-time for Z_0 pfaffian only
  double m_cpu_pfaffian;		// cpu-time for triplets only
  double m_cpu_ext;		// cpu-time for triplets only

	bool m_exact;

  Real m_logZ_exact;      // log partition function for the exact solution
  Real m_logZ_bp;         // log partition function for the bp solution
  Real m_logZ_bp_forney;  // log partition function for the bp solution (forney)
	Real m_logZ_beta;			  // log partition of the reparametereized model
	Real m_Z_tau;        	  // loop expansion of the reparametererized model
	                     	  // m_logZ_exact = log(m_Z_tau) + m_logZ_beta

  Real m_error_bp;      // bp error in marginals
  Real m_error_Zbp;     // bp error in logZ

	bool m_negpot;        // true if at least some negative potential

	std::vector<VProb> m_LR_kernel;    // kernels for Linear Response

  double getCpuBP() { return m_cpu_bp; }
  double getCpuBPForney() { return m_cpu_bp_forney; }
  double getCpuPfaffian1() { return m_cpu_pfaffian1; }
  double getCpuPfaffian() { return m_cpu_pfaffian; }
  Real getLogZBP() { return m_logZ_bp; }
  Real getLogZBPForney() { return m_logZ_bp_forney; }

	bool m_random_init;     // initialize bp messages randomly
	int m_gsl_seed;						// random seed

	// Pfaffian ...
	void toForneyStyle();
  bool doBPForney(Real tol, uint maxiter);
	void doPfaffianSeries( size_t all_comb);
	void planar2Matlab(ExtGraph &g, const ExtDiGraph &dg, int verb, char *filename);
	void rebuildForneyIndex(FVertexD a);
	void makeBiconnected(ExtGraph &gext); 
	void ext2Dot(const ExtGraph &m_gext, const char *filename);
	void makeDirectedDFS(ExtVertexD v, ExtGraph &g, ExtDiGraph &digraph, map<ExtVertexD, bool> &mark);
	void orientPostOrder(BaVertexD u, BaVertexD p, const BasicGraph &gfaces,
											const vector<vector<ExtEdgeD> > &clockwise_efaces,
											const map<BaEdgeD, ExtEdgeD> &map_faces_edges,
											const ExtGraph &gext, ExtDiGraph &tree, map<BaVertexD, bool> &mark);
	void allCombinations(	const int &i,
												const int &m,
												const int &n,
												vector<int> &vcomb,
												ulong &ncomb,
												ulong &total_loops,
												const vector<Real> &vtriplets,
												const vector<vector<int> > &vtriplets_nodes,
												bool sign_correction);
	void pfaffianTerm(const int &pt,
										const vector<int> &vremoved,
										const vector<Real> &vtriplets,
										const vector<vector<int> > &vtriplets_nodes,
										ulong &total_loops,
										bool sign_correction);
	ofstream m_fcorr;

	void outputRes(); 
	Real getLogZExact() const { return Real(m_logZ_exact); }

	//void getDistancesForney(const vector<FVertexD> &triplets, Matrix<int> &mat);

	Real m_ZPfaffian;								// full correction
	Real m_ZPfaffian1;							// single-connected correction

	Real getLogZPfaffian() const { return (m_ZPfaffian != Real(0.)) ? m_logZ_bp + Log( m_ZPfaffian) : Real(0.); }
	Real getLogZPfaffian1() const { return (m_ZPfaffian1 != Real(0.)) ? m_logZ_bp + Log( m_ZPfaffian1) : Real(0.); }
	ForneyGraph m_forney;
	ExtGraph m_gext;
  double m_cpu_bp_forney;
  std::vector<FEdgeD> m_rnd_edges_forney;
	typedef std::pair<Real, ulong> PfaffianTerm;   // < value, num_loops >
	vector<PfaffianTerm> m_contrib_pfaffian;
	vector<bool> m_mark_forney;

	class TripletPair {
		public:
		TripletPair(const int &id1, const int &id2, const Real &value) { m_id1 = id1; m_id2 = id2; m_value = value; }
		int m_id1;
		int m_id2;
		Real m_value;
	};
	struct LessAbsTripletPair : public binary_function<const TripletPair &, const TripletPair &, bool> {
	  bool operator()(const TripletPair &a, const TripletPair &b) { return Abs(a.m_value) > Abs(b.m_value); }
	};
};

// Computation of the Pfaffian in O(n^3) with the exact sign
// According to Galbiati & Maffioli "On the computation of Pfaffians", 1991
template <typename T>
T pfaffian(T c, int n, Matrix<T> *mat)
{
	if (n == 1) {
		T ret( (*mat)(0,1) );
		delete mat;
		return ret;
	}
	else {

		int _2n1 = 2*n-1, _2n2 = 2*n-2;

		// find the element in the last row different than zero
		int k = _2n1;
		while ((*mat)(_2n2,k) == T(0)) {
			k--;
		}
		if (k >= 0) {

			// create space for the matrix<Real> and initialize to zeros
			Matrix<T> *mat2 = new Matrix<T>(_2n2, _2n2, T(0));

			if (k != (_2n1)) {

				// swap k~2n-1 columns and rows
				T aux;
				for (int r = 0; r < 2*n; r++) {
					aux = (*mat)(r,_2n1);
					(*mat)(r,_2n1) = (*mat)(r,k);
					(*mat)(r,k) = aux;
				}
				for (int r = 0; r < 2*n; r++) {
					aux = (*mat)(_2n1,r);
					(*mat)(_2n1,r) = (*mat)(k,r);
					(*mat)(k,r) = aux;
				}
			}

			// fill the new matrix
			for (int i = 0; i < _2n2; i++) 
				for (int j = 0; j < _2n2; j++) 
					(*mat2)(i,j) = ((*mat)(_2n2,_2n1)*(*mat)(i,j) - (*mat)(i,_2n2)*(*mat)(j,_2n1) + (*mat)(j,_2n2)*(*mat)(i,_2n1)) / c;

			// get d (the new c and recursively call pfaffian
			T d = (*mat)(_2n2,_2n1);

			// delete parameter matrix
			delete mat;
			T ret = (k != _2n1) ? -pfaffian(d, n-1, mat2) : pfaffian(d, n-1, mat2);
			return ret;

		} // end if found some k...
	}
	return T(0);
}

// Determinant using characteristic polynomial
template <typename T>
T cpoly(Matrix<T> *mat)
{
	//	local r, s, M, c, f, i, j, k, n;
	int n = (*mat).Rows();

	//	c := A[1, 1], -1;
	vector<T> c(n+1),f(n);
	for (int i = 0; i<n; i++) { c[i] = T(0); f[i] = T(0); }
	c[0] = (*mat)[0][0];
	c[1] = -T(1);

	for (int i = 1; i < n; i++) {

		// M := SubMatrix(A, 1 .. i-1, 1 .. i-1);
		Matrix<T> M(i,i,0);
		for (int mx = 0; mx < i; mx++)
			for (int my = 0; my < i; my++)
				M[mx][my] = (*mat)[mx][my];

		// r := SubMatrix(A, i .. i, 1 .. i-1);
		vector<T> r(i);
		for (int ri = 0; ri < i; ri++)
			r[ri] = (*mat)[i][ri];

		// s := SubMatrix(A, 1 .. i-1, i .. i);
		vector<T> s(i);
		T rs(0.);
		for (int si = 0; si < i; si++) {
			s[si] = (*mat)[si][i];
			rs += r[si]*s[si];
		}
//		cout << "M" <<endl;	for (int mx = 0; mx < i; mx++) {for (int my = 0; my < i; my++)cout << M[mx][my] << "\t";cout << endl;}
//		cout << "r" <<endl;	for (int ri = 0; ri < i; ri++)cout << r[ri] << " ";	cout << endl;

		// f := -1, A[i, i], Multiply(r, s)[1, 1];
		vector<T> f(3+i);
//	cout << endl << "f is " << 3+i << endl;
		
		f[0] = -T(1);
		f[1] = (*mat)[i][i];
		f[2] = rs;

		vector<T> rM(i);
		for (int rMi = 0; rMi < i; rMi++)
			rM[rMi] = r[rMi];

		for (int j = 0; j < i-1; j++) {

			vector<T> rM2(i,T(0));
			// r := Multiply(r, M);
			// r accumulates rM^(j+1)
			for (int my = 0; my < i; my++) {
				for (int rMi = 0; rMi < i; rMi++) {
					rM2[my] += rM[rMi] * M[rMi][my];
				}
			}
			for (int rMi = 0; rMi < i; rMi++)
				rM[rMi] = rM2[rMi];

//cout << "rM" <<endl;for (int ri = 0; ri < i; ri++)cout << rM[ri] << " ";cout << endl;

			// f := f, Multiply(r, s)[1, 1]

			for (int si =0; si < i; si++) {
				f[3+j] += s[si]*rM[si];
			}

		}
//		cout << "f : "; for (int ii = 0; ii < (3+i-1); ii++)cout << f[ii] << " ";cout << endl;

		//	c := seq(   sum(c[k]*f[1+k-j], k = max(1, j) .. i),     j = 0 .. i)
		vector<T> c2(i+2, T(0));

		// first elem
		for (int k = 0; k <= i; k++) {
			c2[0] += c[k] * f[1+k];
		}

		// next ones
		for (int j = 0; j <= i; j++) {
			for (int k = j; k <= i; k++) {
//cout << "accessing f[" << k-j << "]" << endl;
				c2[j+1] += c[k] * f[k-j];
			}
		}
		for (int ci = 0; ci < i+2; ci++) 
			c[ci] = c2[ci];
		//		cout << "---------------------" << endl;
	}

	return c[0];
}

// Computation of the Pfaffian in O(n^3) with the exact sign
// According to Galbiati & Maffioli "On the computation of Pfaffians", 1991
template <typename T>
T pfaffian2(T c, int n, Matrix<T> *mat)
{
//		cout << "in pfaffian2 " << n << endl;
//	//	cout << setprecision(2); cout.setf(ios::fixed);
//		cout.setf(ios::scientific);
//		cout << "c = " << c << " mat =" << endl << (*mat) << endl;
//		cout << "-----------------------------------------------------------" << endl;

	if (n == 1) {
		T ret( (*mat)(0,1) );
		delete mat;
		return ret;
	}
	else {

//cout << "check if swapping "; cout.flush();
		int _2n1 = 2*n-1, _2n2 = 2*n-2;

		// find the element in the last row different than zero
		int k = _2n1;
		while (((*mat)(_2n2,k) == T(0)) && k) {
//cout << (*mat)(_2n2,k) << "\t" << k; cout.flush();
			k--;
		}
//cout << "k is " << k << endl;
		if (k >= 0) {

			// create space for the matrix<Real> and initialize to zeros
			Matrix<T> *mat2 = new Matrix<T>(_2n2, _2n2, T(0));
//cout << "created " << endl;
			if (k != (_2n1)) {

				// swap k~2n-1 columns and rows
				T aux;
				for (int r = 0; r < 2*n; r++) {
					aux = (*mat)(r,_2n1);
					(*mat)(r,_2n1) = (*mat)(r,k);
					(*mat)(r,k) = aux;
				}
				for (int r = 0; r < 2*n; r++) {
					aux = (*mat)(_2n1,r);
					(*mat)(_2n1,r) = (*mat)(k,r);
					(*mat)(k,r) = aux;
				}
			}

			// fill the new matrix
			for (int i = 0; i < _2n2; i++) 
				for (int j = 0; j < _2n2; j++) 
					(*mat2)(i,j) = ((*mat)(_2n2,_2n1)*(*mat)(i,j) - (*mat)(i,_2n2)*(*mat)(j,_2n1) + (*mat)(j,_2n2)*(*mat)(i,_2n1)) / c;

			// get d (the new c and recursively call pfaffian
			// TODO: delete allocated memory
			T d = (*mat)(_2n2,_2n1);

			// delete parameter matrix
			delete mat;
//cout << "-----------------------------------------------------------" << endl;
			T ret = (k != _2n1) ? -pfaffian2(d, n-1, mat2) : pfaffian2(d, n-1, mat2);

			return ret;

		} // end if found some k...
	}
	return T(0);
}



#endif	    /* !FGRAPH_HH_ */
