/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    and Bastian Wemmenhove [b dot wemmenhove at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands

    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __GMR_H__
#define __GMR_H__


#include "factorgraph.h"
#include "daialg.h"

using namespace std;

class GMR : public DAIAlgFG {
 protected:
  typedef struct { size_t i; size_t j; size_t I; } _ijI_type; // see lcbp.h
  typedef struct { size_t i; size_t I; size_t J; } _iIJ_type;
  typedef struct { size_t i; size_t j; size_t J; } _ijJ_type;
  typedef struct { size_t j; size_t i; size_t k; } _jik_type;

  vector<Factor> _Pmess;                  // see bp.h
  vector<Factor> _mumess;
  vector<Factor> _lambdamess;
  vector<Factor> _rhomess;
  vector<Factor> _numess;
  vector<Factor> _concorrel;
  
  vector<Factor> _beliefs;

  vector<_ijI_type> _ijI;                // (see lcbp)

  std::vector<std::vector<size_t> > _EijI;

  vector<_iIJ_type> _iIJ;         

  std::vector<std::vector<size_t> > _EiIJ;

  vector<_ijJ_type> _ijJ;

  std::vector<std::vector<size_t> > _EijJ;

  vector<_jik_type> _jik;
 
  std::vector<std::vector<size_t> > _Ejik;

 public:
  // "default constructor" ??
  GMR() : DAIAlgFG() {};
    // other constructor, using the factorgraph. this construction is the
    // same as in bp.h 
    GMR* clone() const { return new GMR(*this); }
    GMR* create() const { return new GMR(); }

    GMR(const FactorGraph & fg, const Properties & opts) : DAIAlgFG(fg,opts){
      assert( checkProperties() );
      Regenerate();
    }
      static const char *Name;

  /// Sets epsilon
  void Eps( double epsilon ) { SetProperty("epsilon", epsilon); }
  /// Gets epsilon
  double Eps() const { return GetPropertyAs<double>("epsilon"); }

  size_t nr_ijI() const { return _ijI.size(); };

  size_t nr_iIJ() const { return _iIJ.size(); };

  size_t nr_ijJ() const { return _ijJ.size(); };

  size_t nr_jik() const { return _jik.size(); };

  _ijI_type & Pedge(size_t ind) { return _ijI[ind]; }
  std::vector<_ijI_type> & Pedges() { return _ijI; }
  _iIJ_type & Rhoedge(size_t ind) { return _iIJ[ind]; }
  std::vector<_iIJ_type> & Rhoedges() { return _iIJ; }
  _ijJ_type & Nedge(size_t ind) { return _ijJ[ind]; }
  std::vector<_ijJ_type> & Nedges() { return _ijJ; }
  _jik_type & Corpair(size_t ind) { return _jik[ind]; }
  std::vector<_jik_type> & Cedges() { return _jik; }

  size_t VVV2ijI( const size_t i1, const size_t i2, const size_t I) const {
    return _EijI[i1][i2]; }
  size_t VVV2iIJ( const size_t i1, const size_t I, const size_t J) const {
    return _EiIJ[VV2E(i1,I)][J]; }
  size_t VVV2ijJ( const size_t i1, const size_t i2, const size_t J) const {
    return _EijJ[_EijI[i1][i2]][J]; }
  size_t VVV2jik( const size_t i1, const size_t i2, const size_t i3) const {
    return _Ejik[_EijI[i2][i1]][i3]; }

  Factor & Pmes(size_t i1, size_t i2) { return (_Pmess[_EijI[i1][i2]]); }
  Factor & Mumes(size_t i1, size_t i2) { return (_mumess[VV2E(i1,i2)] ); }
  const Factor & Mumes(size_t i1, size_t i2) const { return (_mumess[VV2E(i1,i2)] ); }
  Factor & Lammes(size_t i1, size_t i2) { return (_lambdamess[VV2E(i1,i2)] ); }
  const Factor & Lammes(size_t i1, size_t i2) const { return (_lambdamess[VV2E(i1,i2)] ); }
  Factor & Rhomes(size_t i, size_t I, size_t J){ 
    return (_rhomess[_EiIJ[VV2E(i,I)][J]] ) ; }           
  const Factor & Rhomes(size_t i, size_t I, size_t J) const {
      return (_rhomess[_EiIJ[VV2E(i,I)][J]] ) ; }
  Factor & Numes(size_t i1, size_t i2, size_t J){   
    return (_numess[_EijJ[_EijI[i1][i2]][J]] ) ; }
  const Factor & Numes(size_t i1, size_t i2, size_t J) const {
      return (_numess[_EijJ[_EijI[i1][i2]][J]] ) ; }
  Factor & Concor(size_t i1, size_t i2, size_t i3){
    return (_concorrel[_Ejik[_EijI[i1][i2]][i3]] ) ; }
  const Factor & Concor(size_t i1, size_t i2, size_t i3) const {
    return (_concorrel[_Ejik[_EijI[i1][i2]][i3]] ) ; }    
  
  void Regenerate();
  double run();
  void UpdateOtherMessages();
  
  void UpdateMumes(size_t iI);
  
  void UpdateLammes(size_t iI);
  
  void UpdateRhomes(size_t iIJ);
  
  void UpdateNumes(size_t ijJ);
  
  void UpdatePmess(size_t ijI);
  void UpdatePmessNew( size_t ijI);
  void Initcavs( const string & name, const Properties & opts);

  vector<Factor> calcConcorrel( const InfAlg & obj, const VarSet& ns, bool reInit );
  void init();
  Factor belief1( size_t i ) const;
  void calcbeliefs();
  Factor belief (const Var &n) const;
  Factor belief (const VarSet &n) const { assert( 0 == 1 ); }
  vector<Factor> beliefs() const { return _beliefs; }
  string identify() const ;
  bool checkProperties();
  bool reInit() const { return GetPropertyAs<bool>("reinit"); }
  Complex logZ() const { return NAN; }


};


// Use a lcbp function to calculate the correlation coefficients



// then propagate updates until convergence, looking very much like 
// bp stuff

// The classes to inherit from are at least Factor (better: Daialg) 
// and perhaps lcbp

// We need a number of private functions:

// 1. the P^{(i)}(x_j), which will be something like messages, but they will
// not be from factor to node, since they explicitly go from node to node.

// 2. the mu_{\alpha \to i}(x_i), which will be the same as the messages

// 3. the \lambda_{\alpha \to i}(x_i), having same representation as the
// messages

// 4. the \rho_{(\alpha,\beta)\to i}(x_i) messages, that have quite a 
// difficult representation, since they involve more than one neighbouring
// factor if i

// 5. the \nu_{\alpha \to i}(x_i,x_j) which depends on factor alpha and 
// node j, both of which are in the cavity of i.


#endif
