/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <algorithm>
#include <map>
#include <set>
#include "lcbp1.h"
#include "diffs.h"
#include "util.h"
#include "alldai.h"


using namespace std;


const char *LCBP1::Name = "LCBP1";


bool LCBP1::checkProperties() {
    if( !HasProperty("type") )
        return false;
    if( !HasProperty("cavity") )
        return false;
    if( !HasProperty("updates") )
        return false;
    if( !HasProperty("tol") )
        return false;
    if (!HasProperty("maxiter") )
        return false;
    if (!HasProperty("verbose") )
        return false;

    _properties.type = FromStringTo<TypeType>("type");
    _properties.updates = FromStringTo<UpdateType>("updates");
    _properties.cavity = FromStringTo<CavityType>("cavity");
    ConvertPropertyTo<double>("tol");
    ConvertPropertyTo<size_t>("maxiter");
    ConvertPropertyTo<size_t>("verbose");

    if (HasProperty("cavainame") )
        ConvertPropertyTo<string>("cavainame");
    if (HasProperty("cavaiopts") )
        ConvertPropertyTo<Properties>("cavaiopts");
    if( HasProperty("reinit") )
        _properties.reinit = FromStringTo<bool>("reinit");
    
    return true;
}


LCBP1::LCBP1(const FactorGraph & fg, const Properties &opts) : DAIAlgFG(fg, opts) {
    assert( checkProperties() );

    // calc ijI
    for( size_t i=0; i < nrVars(); i++ ) {
        for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ )
            if( Type() == TypeType::HOIPLUS || Type() == TypeType::HOIMINUS ) {
                _ijI_type _ijI_entry;
                _ijI_entry.i = i;
                _ijI_entry.j = -1U;
                _ijI_entry.I = *I;

                _ijI.push_back(_ijI_entry);
            } else
                for( _nb_cit j = nbF(*I).begin(); j != nbF(*I).end(); j++ ) 
                    if( *j != i ) {
                        _ijI_type _ijI_entry;
                        _ijI_entry.i = i;
                        _ijI_entry.j = *j;
                        _ijI_entry.I = *I;

                        _ijI.push_back(_ijI_entry);
                    }
    }

    // create pancakes
    _pancakes.resize(nrVars());
   
    // create cavitydists
    for( size_t i=0; i < nrVars(); i++ )
        _cavitydists.push_back(Factor(delta(var(i))));

    // create phis
    _phis.reserve(nrEdges());
    for( size_t iI = 0; iI < nrEdges(); iI++ ) {
        size_t i = edge(iI).first;
        size_t I = edge(iI).second;
        _phis.push_back( Factor( factor(I).vars() / var(i) ) );
    }

    // create beliefs
    for( size_t i=0; i < nrVars(); i++ )
        _beliefs.push_back(Factor(var(i)));
}


string LCBP1::identify() const { 
    stringstream result (stringstream::out);
    result << Name << GetProperties();
    return result.str();
}


void LCBP1::CalcBelief (size_t i) {
    _beliefs[i] = _pancakes[i].marginal(var(i));
}


double LCBP1::CalcCavityDist (size_t i, const string &name, const Properties &opts) {
    Factor Bi;
    double maxdiff = 0;

    if( Verbose() >= 2 )
        cout << "Initing cavity " << var(i) << "(" << delta(var(i)).size() << " vars, " << delta(var(i)).stateSpace() << " states)" << endl;

    if( Cavity() == CavityType::UNIFORM )
        Bi = Factor(delta(var(i)));
    else {
        InfAlg *cav = newInfAlg( name, *this, opts );
        cav->makeCavity( var(i) );

        if( Cavity() == CavityType::FULL )
            Bi = calcMarginal( *cav, cav->delta(var(i)), reInit() );
        else if( Cavity() == CavityType::PAIR )
            Bi = calcMarginal2ndO( *cav, cav->delta(var(i)), reInit() );
        else if( Cavity() == CavityType::PAIR2 ) {
            vector<Factor> pairbeliefs = calcPairBeliefsNew( *cav, cav->delta(var(i)), reInit() );
            for( size_t ij = 0; ij < pairbeliefs.size(); ij++ )
                Bi *= pairbeliefs[ij];
        }
        maxdiff = cav->MaxDiff();
        delete cav;
    }
    Bi.normalize( _normtype );
    _cavitydists[i] = Bi;

    return maxdiff;
}


double LCBP1::InitCavityDists (const string &name, const Properties &opts) {
    double tic = toc();

    if( Verbose() >= 1 ) {
        cout << "LCBP1::InitCavityDists:  ";
        if( Cavity() == CavityType::UNIFORM )
            cout << "Using uniform initial cavity distributions" << endl;
        else if( Cavity() == CavityType::FULL )
            cout << "Using full " << name << opts << "...";
        else if( Cavity() == CavityType::PAIR )
            cout << "Using pairwise " << name << opts << "...";
        else if( Cavity() == CavityType::PAIR2 )
            cout << "Using pairwise(new) " << name << opts << "...";
    }

    double maxdiff = 0.0;
    for( size_t i = 0; i < nrVars(); i++ ) {
        double md = CalcCavityDist(i, name, opts);
        if( md > maxdiff )
            maxdiff = md;
    }
    init();

    if( Verbose() >= 1 ) {
        cout << "used " << toc() - tic << " clocks." << endl;
    }

    return maxdiff;
}


long LCBP1::SetCavityDists (vector<Factor> &Q) {
    if( Verbose() >= 1 ) 
        cout << "LCBP1::SetCavityDists:  Setting initial cavity distributions" << endl;
    if( Q.size() != nrVars() )
        return -1;
    for( size_t i = 0; i < nrVars(); i++ ) {
        if( _cavitydists[i].vars() != Q[i].vars() ) {
            return i+1;
        } else
            _cavitydists[i] = Q[i];
    }
    init();
    return 0;
}


void LCBP1::init() {
    for( size_t iI = 0; iI < nrEdges(); iI++ ) {
        if( (Updates() == UpdateType::SEQRND) && (Type() == TypeType::HOIPLUS || Type() == TypeType::HOIMINUS) )
//            for( size_t pos = 0; pos < _phis[iI].p().size(); pos++ )
//                _phis[iI].p()[pos] = rnd_uniform();
            _phis[iI].randomize();
        else
            _phis[iI].fill(1.0);
    }
    for( size_t i = 0; i < nrVars(); i++ ) {
        _pancakes[i] = _cavitydists[i];
        
        for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ ) {
            _pancakes[i] *= factor(*I);
            if( (Updates() == UpdateType::SEQRND) && (Type() == TypeType::MINUS || Type() == TypeType::PLUS || Type() == TypeType::STAR) ) {
                for( _nb_cit j = nbF(*I).begin(); j != nbF(*I).end(); j++ ) 
                    if( *j != i ) {
                        Factor j_noise(var(*j));
                        j_noise.randomize();
                        _pancakes[i] *= j_noise;
                    }
            } else if( (Updates() == UpdateType::SEQRND) && (Type() == TypeType::HOIPLUS || Type() == TypeType::HOIMINUS) ) {
              _pancakes[i] *= _phis[VV2E(i,*I)];
//              cout << "rnd-izing with " << _phis[VV2E(i,*I)] << endl;
            }
        }
        
        _pancakes[i].normalize( _normtype );

        CalcBelief(i);
    }
}


Factor LCBP1::NewPancake (size_t ijI, bool & hasNaNs) {
    size_t i = _ijI[ijI].i;
    size_t j = _ijI[ijI].j;
    size_t I = _ijI[ijI].I;

    Factor piet = _pancakes[i];

    // recalculate _pancake[i]
    
    if( Type() == TypeType::PLUS ) {
        Factor quot = _beliefs[j].divided_by(_pancakes[i].partSum(var(j)));
        piet *= quot;
//        _phis[ijI] *= quot;
    }
    else if( Type() == TypeType::MINUS ) {
        Factor numer = _pancakes[j] * factor(I).inverse();
        Factor denom = _pancakes[i] * factor(I).inverse();
        Factor quot = numer.partSum(var(j)).divided_by(denom.partSum(var(j)));
        piet *= quot;
//        _phis[ijI] *= quot;
    } else if( Type() == TypeType::STAR ) {
/* THIS UPDATE EQUATION (LCBP1^*) IS VERY BAD FOR NON-PAIRWISE INTERACTIONS
 * BUT SEEMS TO BE BETTER THAN LCBP1^+ AND LCBP1^- FOR PAIRWISE INTERACTIONS,
 * ALTHOUGH CONVERGENCE CAN BE A PROBLEM LIKE FOR LCBP1^-
 */
        size_t iI = VV2E(i, I);
		VarSet Ivars = factor(I).vars();
        Factor A_I  = _pancakes[j].partSum( Ivars );
        Factor A_Ii = (_pancakes[i] * _phis[iI].inverse()).partSum( Ivars );
        Factor quot = (A_I.divided_by(A_Ii)).marginal( Ivars / var(i) );

/* THIS SLIGHT VARIATION GIVES BAD RESULTS FOR NON-PAIRWISE INTERACTIONS
 * BUT GOOD RESULTS FOR PAIRWISE INTERACTIONS (STRANGE)
 */
/*        size_t iI = VV2E(i, I);
		VarSet Ivars = factor(I).vars();
        Factor A_I;
        for( VarSet::const_iterator k = Ivars.begin(); k != Ivars.end(); k++ )
            if( var(i) != *k )
                A_I *= (_pancakes[findVar(*k)]).partSum( Ivars );
        A_I ^= (1.0 / (Ivars.size() - 1));
        Factor A_Ii = (_pancakes[i] * _phis[iI].inverse()).partSum( Ivars );
        Factor quot = (A_I.divided_by(A_Ii)).marginal( Ivars / var(i) );
*/

        piet *= quot.divided_by( _phis[iI] ).normalized( _normtype );
        _phis[iI] = quot.normalized( _normtype );
    } else if( Type() == TypeType::HOIMINUS ) {
/* THIS IMPROVES OVER LCBP1+ AND LCBP1- FOR NON-PAIRWISE INTERACTIONS
 * IT MODIFIES LCBP1-, I.E. THE CONNECTING FACTOR IS ABSENT
 */
        size_t iI = VV2E(i, I);
		VarSet Ivars = factor(I).vars();
        Factor A_I;
        for( VarSet::const_iterator k = Ivars.begin(); k != Ivars.end(); k++ )
            if( var(i) != *k )
                A_I *= (_pancakes[findVar(*k)] * factor(I).inverse()).partSum( Ivars / var(i) );
        A_I ^= (1.0 / (Ivars.size() - 1));
        Factor A_Ii = (_pancakes[i] * factor(I).inverse() * _phis[iI].inverse()).partSum( Ivars / var(i) );
        Factor quot = A_I.divided_by(A_Ii);

        piet *= quot.divided_by( _phis[iI] ).normalized( _normtype );
        _phis[iI] = quot.normalized( _normtype );
    } else if( Type() == TypeType::HOIPLUS ) {
/* THIS IMPROVES OVER LCBP1+ AND LCBP1- FOR NON-PAIRWISE INTERACTIONS
 * IT MODIFIES LCBP1+, I.E. THE CONNECTING FACTOR IS PRESENT
 */
        size_t iI = VV2E(i, I);
		VarSet Ivars = factor(I).vars();
        Factor A_I;
        for( VarSet::const_iterator k = Ivars.begin(); k != Ivars.end(); k++ )
            if( var(i) != *k )
                A_I *= (_pancakes[findVar(*k)]).partSum( Ivars / var(i) );
        A_I ^= (1.0 / (Ivars.size() - 1));
        Factor A_Ii = (_pancakes[i] * _phis[iI].inverse()).partSum( Ivars / var(i) );
        Factor quot = A_I.divided_by(A_Ii);

/* THIS IS YET ANOTHER VARIATION, WHAT DOES IT DO?
 */
/*        size_t iI = VV2E(i, I);
		VarSet Ivars = factor(I).vars();
        Factor A_I;
        for( VarSet::const_iterator k = Ivars.begin(); k != Ivars.end(); k++ )
            if( var(i) != *k )
                A_I *= (_pancakes[findVar(*k)] * factor(I).inverse()).partSum( Ivars / var(i) );
        Factor A_Ii = (_pancakes[i] * factor(I).inverse() * _phis[iI].inverse()).partSum( Ivars /var(i) );
        A_Ii ^= (Ivars.size() - 1);
        Factor quot = ((A_I.divided_by(A_Ii)).marginal(Ivars / var(i)))^(1 / (Ivars.size() - 1));
*/
/*        size_t iI = VV2E(i, I);
		VarSet Ivars = factor(I).vars();
        Factor A_Ij = (_pancakes[j] * factor(I).inverse()).partSum( Ivars );
        Factor A_Ii = (_pancakes[i] * factor(I).inverse() * _phis[iI].inverse()).partSum( Ivars );
        Factor quot = (A_Ij.divided_by(A_Ii)).marginal( Ivars / var(i) );
*/
/*        Factor factquot;
        for( VarSet::const_iterator k = Ivars.begin(); k != Ivars.end(); k++ )
            if( var(i) != *k )
                factquot *= quot.marginal( *k );
        quot = factquot.normalized( _normtype );
*/

        piet *= quot.divided_by( _phis[iI] ).normalized( _normtype );
        _phis[iI] = quot.normalized( _normtype );
        
/*      THIS IMPROVES OVER LCBP1+ and LCBP1- FOR PAIRWISE INTERACTIONS,
        BUT NOT FOR HIGHER ORDER INTERACTIONS

        size_t jiI = ijI_index(j,i,I);
		VarSet Ivars = factor(I).vars();
		Factor A_Ij = (_pancakes[j] * factor(I).inverse() * _phis[jiI].inverse()).partSum( Ivars );
		Factor A_Ii = (_pancakes[i] * factor(I).inverse() * _phis[ijI].inverse()).partSum( Ivars );
		Factor quot = (A_Ij.divided_by(A_Ii) * _phis[jiI]).marginal(var(j));
		piet *= quot.divided_by(_phis[ijI]);
        _phis[ijI] = quot.normalize( _normtype );
*/

        
/*      OLD ATTEMPT FOR IMPROVEMENT, DID NOT WORK

        VarSet jset = factor(I).vars() / var(i);
        Factor jjoint;
        for( VarSet::const_iterator _j = jset.begin(); _j != jset.end(); _j++ ) {
            size_t jind = findVar(*_j);
            jjoint *= (pancake(jind) * factor(I).inverse()).marginal(jset);
        }
        jjoint ^= 1.0 / jset.size();
//      piet *= jjoint.marginal(var(j)).divided_by(_pancakes[i].partSum(var(j)));
        piet *= jjoint.divided_by((_pancakes[i] * factor(I).inverse()).partSum(jset));
*/
    }

    piet.normalize( _normtype );

    if( piet.hasNaNs() ) {
        cout << "LCBP1::NewPancake(" << ijI << "):  has NaNs!" << endl;
        hasNaNs = true;
    }

    return piet;
}


double LCBP1::run() {
    if( Verbose() >= 1 )
        cout << "Starting " << identify() << "...";
    if( Verbose() >= 2 )
        cout << endl;

    double tic = toc();
    Diffs diffs(nrVars(), 1.0);

    updateMaxDiff( InitCavityDists(GetPropertyAs<string>("cavainame"), GetPropertyAs<Properties>("cavaiopts")) );

    vector<Factor> old_beliefs;
    for(size_t i=0; i < nrVars(); i++ )
        old_beliefs.push_back(belief(i));

    bool hasNaNs = false;
    for( size_t i=0; i < nrVars(); i++ )
        if( _pancakes[i].hasNaNs() ) {
            hasNaNs = true;
            break;
        }
    if( hasNaNs ) {
        cout << "LCBP1::run:  initial _pancakes has NaNs!" << endl;
        return NAN;
    }

    vector<long> update_seq(nr_ijI(),0);
    for( size_t k=0; k < nr_ijI(); k++ )
        update_seq[k] = k;

    size_t iter=0;

    // do several passes over the network until maximum number of iterations has
    // been reached or until the maximum belief difference is smaller than tolerance
    for( iter=0; iter < MaxIter() && diffs.max() > Tol(); iter++ ) {
        // Sequential updates
        if( Updates() == UpdateType::SEQRND )
            random_shuffle( update_seq.begin(), update_seq.end() );
        
        for( size_t t=0; t < nr_ijI(); t++ ) {
            long ijI = update_seq[t];
            long i = _ijI[ijI].i;
            _pancakes[i] = NewPancake(ijI, hasNaNs);
            if( hasNaNs )
                return NAN;
            CalcBelief(i);
        }

        // compare new beliefs with old ones
        for(size_t i=0; i < nrVars(); i++ ) {
            diffs.push( dist( belief(i), old_beliefs[i], Prob::DISTLINF ) );
            old_beliefs[i] = belief(i);
        }

        if( Verbose() >= 3 )
            cout << "LCBP1::run:  maxdiff " << diffs.max() << " after " << iter+1 << " passes" << endl;
    }

    Iterations( iter );
    updateMaxDiff( diffs.max() );

    if( Verbose() >= 1 ) {
        if( diffs.max() > Tol() ) {
            if( Verbose() == 1 )
                cout << endl;
                cout << "LCBP1::run:  WARNING: not converged within " << MaxIter() << " passes (" << toc() - tic << " clocks)...final maxdiff:" << diffs.max() << endl;
        } else {
            if( Verbose() >= 2 )
                cout << "LCBP1::run:  ";
                cout << "converged in " << iter << " passes (" << toc() - tic << " clocks)." << endl;
        }
    }

    return diffs.max();
}


size_t LCBP1::ijI_index(size_t i, size_t j, size_t I) const {
    size_t ind = 0;
    for( ; ind < _ijI.size(); ind++ )
        if( (_ijI[ind].i == i) && (_ijI[ind].I == I) && (_ijI[ind].j == j) )
            break;
    assert( ind != _ijI.size() );
    return ind;
}
