/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands

    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __LCBP1_H__
#define __LCBP1_H__


#include "daialg.h"
#include "enum.h"
#include "factorgraph.h"


using namespace std;


class LCBP1 : public DAIAlgFG {
    protected:
        typedef struct { size_t i; size_t j; size_t I; } _ijI_type;

        vector<Factor>      _pancakes;      // used by all LCBP1 types (psi_I is stored in the pancake)
        vector<Factor>      _cavitydists;   // used by all LCBP1 types to store the approximate cavity distribution
        /// _phis[VV2E(i,I)] corresponds to \f$ \phi^{\setminus i}_I(x_{I \setminus i}) \f$
        vector<Factor>      _phis;

        /// Single variable beliefs
        vector<Factor>      _beliefs;

        /// For each pair (i,j) with j in delta(i), store i, j and the common factor I
        vector<_ijI_type>   _ijI;

    public:
        ENUM(TypeType,PLUS,MINUS,STAR,HOIMINUS,HOIPLUS)
        ENUM(CavityType,FULL,PAIR,PAIR2,UNIFORM)
        ENUM(UpdateType,SEQFIX,SEQRND,NONE)

        struct {
            TypeType   type;
            UpdateType updates;
            CavityType cavity;
            bool       reinit;
        } _properties;
        TypeType Type() const { return _properties.type; }
        CavityType Cavity() const { return _properties.cavity; }
        UpdateType Updates() const { return _properties.updates; }
        bool reInit() const { return _properties.reinit; }
        
        /// Default constructor
        LCBP1() : DAIAlgFG() {};
        /// Copy constructor
        LCBP1(const LCBP1 & x) : DAIAlgFG(x), _pancakes(x._pancakes), _cavitydists(x._cavitydists), _phis(x._phis), _beliefs(x._beliefs), _ijI(x._ijI) {};
        /// Clone function
        LCBP1* clone() const { return new LCBP1(*this); }
        /// Create function
        LCBP1* create() const { return new LCBP1(); }
        /// Construct LCBP1 object from a FactorGraph and parameters
        LCBP1(const FactorGraph & fg, const Properties &opts);
        /// Assignment operator
        LCBP1& operator=(const LCBP1 & x) {
            if( this != &x ) {
                DAIAlgFG::operator=(x);
                _pancakes       = x._pancakes;
                _cavitydists    = x._cavitydists;
                _phis           = x._phis;
                _beliefs        = x._beliefs;
                _ijI            = x._ijI;
            }
            return *this;
        }

        static const char *Name;
        double CalcCavityDist (size_t i, const string &name, const Properties &opts);
        double InitCavityDists (const string &name, const Properties &opts);
        long SetCavityDists (vector<Factor> &Q);

        void init();
        Factor NewPancake (size_t ijI, bool & hasNaNs);
        double run();

        string identify() const;
        Factor belief (const Var &n) const { return( _beliefs[findVar(n)] ); }
        Factor belief (const VarSet &ns) const { assert( 0 == 1 ); }
        vector<Factor> beliefs() const { return _beliefs; }
        Complex logZ() const { return NAN; }
        void CalcBelief (size_t i);
        const Factor & belief (size_t i) const { return _beliefs[i]; };
        const Factor & pancake (size_t i) const { return _pancakes[i]; };
        const Factor & cavitydist (size_t i) const { return _cavitydists[i]; };
        size_t nr_ijI() const { return _ijI.size(); };
        size_t ijI_index(size_t i, size_t j, size_t I) const;

        void clamp( const Var &n, size_t i ) { assert( 0 == 1 ); }
        void undoProbs( const VarSet &ns ) { assert( 0 == 1 ); }
        void saveProbs( const VarSet &ns ) { assert( 0 == 1 ); }
        void makeFactorCavity(size_t I) { assert( 0 == 1 ); }
        virtual void makeCavity(const Var & n) { assert( 0 == 1 ); }
        bool checkProperties();
};


#endif
