/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <algorithm>
#include <map>
#include <set>
#include "diffs.h"
#include "util.h"
#include "alldai.h"


using namespace std;


const char *LCBP2::Name = "LCBP2";


bool LCBP2::checkProperties() {
    if( !HasProperty("cavity") )
        return false;
    if( !HasProperty("updates") )
        return false;
    if( !HasProperty("tol") )
        return false;
    if (!HasProperty("maxiter") )
        return false;
    if (!HasProperty("verbose") )
        return false;
    
    _properties.updates = FromStringTo<UpdateType>("updates");
    _properties.cavity = FromStringTo<CavityType>("cavity");
    ConvertPropertyTo<double>("tol");
    ConvertPropertyTo<size_t>("maxiter");
    ConvertPropertyTo<size_t>("verbose");

    if( HasProperty("reinit") )
        _properties.reinit = FromStringTo<bool>("reinit");
    if (HasProperty("cavainame") )
        ConvertPropertyTo<string>("cavainame");
    if (HasProperty("cavaiopts") )
        ConvertPropertyTo<Properties>("cavaiopts");

    return true;
}


LCBP2::LCBP2(const FactorGraph & fg, const Properties &opts) : DAIAlgFG(fg, opts) {
    assert( checkProperties() );

    // calc ij
    for( size_t i=0; i < nrVars(); i++ ) {
        _nb_t di;
        for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ )
            for( _nb_cit j = nbF(*I).begin(); j != nbF(*I).end(); j++ ) 
                if( *j != i )
                    di.push_back(*j);
        // Remove duplicates from di
        _nb_t::iterator new_end = unique(di.begin(), di.end());
        di.erase( new_end, di.end() );
        // Caclulate ij
        for( _nb_t::iterator j = di.begin(); j != di.end(); j++ ) {
            _ij_type ij_entry;
            ij_entry.i = i;
            ij_entry.j = *j;
            _ij.push_back(ij_entry);
        }
    }

    // create beliefs
    for( size_t i=0; i < nrVars(); i++ )
        _beliefs.push_back(Factor(var(i)));

    // create pancakes
    _pancakes.resize(nrVars(), NULL);

    // create cavitydists
    _cavitydists.resize(nrVars());
}


string LCBP2::identify() const { 
    stringstream result (stringstream::out);
    result << Name << GetProperties();
    return result.str();
}


void LCBP2::CalcBelief (size_t i) {
    _pancakes[i]->init();       // FIXME
    _pancakes[i]->run();
    _beliefs[i] = _pancakes[i]->belief(var(i));
}


double LCBP2::CalcCavityDist (size_t i, const string &name, const Properties &opts) {
    double maxdiff = 0;

    if( Verbose() >= 2 )
        cout << "Initing cavity " << var(i) << "(" << delta(var(i)).size() << " vars, " << delta(var(i)).stateSpace() << " states)" << endl;

    if( Cavity() == CavityType::UNIFORM ) {
        vector<Factor> pairbeliefs;

        VarSet ns = delta(var(i));
        pairbeliefs.reserve( ns.size() * (ns.size() - 1) / 2 );
        
        VarSet::const_iterator nj = ns.begin();
        for( VarSet::const_iterator nj = ns.begin(); nj != ns.end(); nj++ )
            for( VarSet::const_iterator nk = nj; (++nk) != ns.end(); )
                pairbeliefs.push_back( Factor(*nj | *nk) );

        _cavitydists[i] = pairbeliefs;
        maxdiff = 0.0;
    } else if( Cavity() == CavityType::PAIR ) {
        InfAlg *cav = newInfAlg( name, *this, opts );
        cav->makeCavity( var(i) );
        _cavitydists[i] = calcPairBeliefs( *cav, delta(var(i)), reInit() );
        // FIXME: INITBP3 used true instead of false above
        maxdiff = cav->MaxDiff();
        delete cav;
    }

    return maxdiff;
}


double LCBP2::InitCavityDists (const string &name, const Properties &opts) {
    double tic = toc();

    if( Verbose() >= 1 )
        cout << "LCBP2::InitCavityDists starting..." << endl;

//    vector<size_t> methods(nrVars(), method);
//    double md = InitCavityDists( methods, tol, maxiter, verbose );
    double maxdiff = 0;
    for( size_t i = 0; i < nrVars(); i++ ) {
        double md = CalcCavityDist(i, name, opts);
        if( md > maxdiff )
            maxdiff = md;
    }
    init();

    if( Verbose() >= 1 )
        cout << "LCBP2::InitCavityDists:  used " << toc() - tic << " clocks." << endl;

    return maxdiff;
}


void LCBP2::init() {
    assert( checkProperties() );

    for( size_t i = 0; i < nrVars(); i++ ) {
        vector<Factor> facs;
        VarSet di = delta(var(i));
        for( VarSet::const_iterator j = di.begin(); j != di.end(); j++ )
            facs.push_back( Factor(*j,1.0) );
        for( size_t I = 0; I < _cavitydists[i].size(); I++ )
            facs.push_back( _cavitydists[i][I] );
        for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ )
            facs.push_back( factor( *I ) );
        FactorGraph fg(facs);

////    stringstream fn (stringstream::out);
////    fn << "/tmp/cavity_" << i << ".fg";
//      char fn[100];
//      sprintf( fn, "/tmp/cavity_%03d.fg", (int)i );
//      fg.WriteToFile(fn);

        /* TREE-EP ON THE STAR CENTRED ON I */
        DEdgeVec tr;
        for( VarSet::const_iterator j = di.begin(); j != di.end(); j++ )
            tr.push_back( DEdge(fg.findVar(var(i)),fg.findVar(*j)) );
        _pancakes[i] = new TreeEP( fg, Properties()("type", string("ORG"))("tree",tr)("tol", 1e-9)("maxiter", 1000UL)("verbose", 0UL) );

        /* JUNCTION-TREE */
//      _pancakes[i] = new JTree( fg, EXACT_JTREE_HUGIN, 0 );

        /* DOUBLE-LOOP CVM ON ALL TRIPLES INCLUDING I */
/*      vector<VarSet> Ivs;
        for( _nb_cit j = _delta[i].begin(); j != _delta[i].end(); j++ ) 
            for( _nb_cit k = j; ++k != _delta[i].end(); )
                Ivs.push_back( var(i) | var(*j) | var(*k) );
        RegionGraph rg(fg, Ivs);
        _pancakes[i] = new HAK( rg, MARGINAL_INIT | HAK_DOUBLELOOP, 1e-9, 1000, 0 );
*/
        /* DOUBLE-LOOP CVM ON CLIQUES */
/*      vector<VarSet> Ivs = fg.Cliques();
        RegionGraph rg(fg, Ivs);
        _pancakes[i] = new HAK( rg, MARGINAL_INIT | HAK_DOUBLELOOP, 1e-9, 1000, 0 );
*/
        /* SIMPLE BP */
//      _pancakes[i] = new BP( FactorGraph(facs), BP_SEQFIXED, 1e-9, 1000, 0 );

        CalcBelief(i);
    }
}


void LCBP2::UpdatePancake( size_t ij ) {
    size_t i = _ij[ij].i;
    size_t j = _ij[ij].j;

    // find index of var j in _pancakes[i]
    size_t j_in_i = _pancakes[i]->findFactor(var(j));
            
    vector<size_t> Is_in_i, Is_in_j;
    
    for( size_t I_in_i = 0; I_in_i < _pancakes[i]->nrFactors(); I_in_i++ )
        if( _pancakes[i]->factor(I_in_i).vars() >> (var(i) | var(j)) ) {
            // factor I_in_i in i contains variables i and j

//          cerr << "var(i): " << var(i) << endl;
//          cerr << "var(j): " << var(j) << endl;
//          cerr << "delta(i): ";
//          for( size_t k = 0; k < _delta[i].size(); k++ )
//              cerr << var(_delta[i][k]) << " ";
//          cerr << endl;
//          cerr << "delta(j): ";
//          for( size_t k = 0; k < _delta[j].size(); k++ )
//              cerr << var(_delta[j][k]) << " ";
//          cerr << endl;
//          cerr << "_pancakes[i]->factor(I_in_i): " << _pancakes[i]->factor(I_in_i).vars() << endl;

            // find the corresponding factor in j
            size_t I_in_j = _pancakes[j]->findFactor( _pancakes[i]->factor(I_in_i).vars() );
            
            Is_in_i.push_back( I_in_i );
            Is_in_j.push_back( I_in_j );
        }
    if( Is_in_i.size() > 1 )
        cout << "LCBP2 != LCBP-II (more connecting factors between " << var(i) << " and " << var(j) << ")" << endl;

    // neutralize all I connecting i and j in _pancakes[i]
    for( size_t I = 0; I < Is_in_i.size(); I++ ) {
        _pancakes[i]->saveProb( Is_in_i[I] );
        _pancakes[i]->makeFactorCavity( Is_in_i[I] );
    }
    // do inference in _pancakes[i]
    _pancakes[i]->init();           // FIXME
    _pancakes[i]->run();
    
    // neutralize all I connecting i and j in _pancakes[j]
    for( size_t I = 0; I < Is_in_j.size(); I++ ) {
        _pancakes[j]->saveProb( Is_in_j[I] );
        _pancakes[j]->makeFactorCavity( Is_in_j[I] );
    }
    // do inference in _pancakes[j]
    _pancakes[j]->init();           // FIXME
    _pancakes[j]->run();

    // do LCBP update
    _pancakes[i]->factor(j_in_i) *= _pancakes[j]->belief(var(j)).divided_by( _pancakes[i]->belief(var(j)) );
    _pancakes[i]->updatedFactor( j_in_i );

    // restore all I connecting i and j in _pancakes[i]
    for( size_t I = 0; I < Is_in_i.size(); I++ )
        _pancakes[i]->undoProb( Is_in_i[I] );

    // restore all I connecting i and j in _pancakes[j]
    for( size_t I = 0; I < Is_in_j.size(); I++ )
        _pancakes[j]->undoProb( Is_in_j[I] );
}


double LCBP2::run() {
    if( Verbose() >= 1 )
        cout << "Starting " << identify() << "...";
    if( Verbose() >= 2 )
        cout << endl;

    double tic = toc();
    Diffs diffs(nrVars(), 1.0);

    updateMaxDiff( InitCavityDists(GetPropertyAs<string>("cavainame"), GetPropertyAs<Properties>("cavaiopts")) );

    vector<Factor> old_beliefs;
    for(size_t i=0; i < nrVars(); i++ )
        old_beliefs.push_back(belief(i));

    vector<long> update_seq(nr_ij(),0);
    for( size_t k=0; k < nr_ij(); k++ )
        update_seq[k] = k;

    size_t iter=0;

    // do several passes over the network until maximum number of iterations has
    // been reached or until the maximum belief difference is smaller than tolerance
    for( iter=0; iter < MaxIter() && diffs.max() > Tol(); iter++ ) {
        // Sequential updates
        if( Updates() == UpdateType::SEQRND )
            random_shuffle( update_seq.begin(), update_seq.end() );
        
        for( size_t t=0; t < nr_ij(); t++ ) {
            long ij = update_seq[t];
            long i = _ij[ij].i;
            UpdatePancake(ij);
            CalcBelief(i);
        }

        // compare new beliefs with old ones
        for(size_t i=0; i < nrVars(); i++ ) {
            diffs.push( dist( belief(i), old_beliefs[i], Prob::DISTLINF ) );
            old_beliefs[i] = belief(i);
        }

        if( Verbose() >= 3 )
            cout << "LCBP2::run:  maxdiff " << diffs.max() << " after " << iter+1 << " passes" << endl;
    }

    Iterations( iter );
    updateMaxDiff( diffs.max() );

    if( Verbose() >= 1 ) {
        if( diffs.max() > Tol() ) {
            if( Verbose() == 1 )
                cout << endl;
                cout << "LCBP2::run:  WARNING: not converged within " << MaxIter() << " passes (" << toc() - tic << " clocks)...final maxdiff:" << diffs.max() << endl;
        } else {
            if( Verbose() >= 2 )
                cout << "LCBP2::run:  ";
                cout << "converged in " << iter << " passes (" << toc() - tic << " clocks)." << endl;
        }
    }

    return diffs.max();
}


Factor LCBP2::belief (const Var &n) const {
    return( _beliefs[findVar( n )] );
}


vector<Factor> LCBP2::beliefs() const {
    return _beliefs;
}
