/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands

    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __LCBP2_H__
#define __LCBP2_H__


#include "daialg.h"
#include "enum.h"
#include "factorgraph.h"


using namespace std;


class LCBP2 : public DAIAlgFG {
    protected:
        typedef struct { size_t i; size_t j; } _ij_type;

        vector<InfAlg*>         _pancakes;
        vector<vector<Factor> > _cavitydists;

        /// Single variable beliefs
        vector<Factor>          _beliefs;

        /// For each pair (i,j) with j in delta(i), store i and j
        vector<_ij_type>        _ij;

    public:
//        ENUM(TypeType,PLUS,MIN,STAR)
        ENUM(CavityType,PAIR,UNIFORM)
        ENUM(UpdateType,SEQFIX,SEQRND,NONE)

        struct {
            UpdateType updates;
            CavityType cavity;
            bool       reinit;
        } _properties;
        CavityType Cavity() const { return _properties.cavity; }
        UpdateType Updates() const { return _properties.updates; }
        bool reInit() const { return _properties.reinit; }
//        TypeType Type() const { return GetPropertyAs<TypeType>("type"); }

        /// Default constructor
        LCBP2() : DAIAlgFG() {};
        /// Copy constructor
        LCBP2(const LCBP2 & x) : DAIAlgFG(x), _cavitydists(x._cavitydists), _beliefs(x._beliefs), _ij(x._ij) {
            // make deep copy of _pancakes
            _pancakes.resize( x._pancakes.size(), NULL );
            for( size_t i = 0; i < x._pancakes.size(); i++ )
                _pancakes[i] = x._pancakes[i]->clone();
        }
        /// Clone function
        LCBP2* clone() const { return new LCBP2(*this); }
        /// Create function
        LCBP2* create() const { return new LCBP2(); }
        /// Construct LCBP2 object from a FactorGraph and parameters
        LCBP2(const FactorGraph & fg, const Properties &opts);
        /// Assignment operator
        LCBP2& operator=(const LCBP2 & x) {
            if( this != &x ) {
                DAIAlgFG::operator=(x);
                // make deep copy of _pancakes
                _pancakes.resize( x._pancakes.size(), NULL );
                for( size_t i = 0; i < x._pancakes.size(); i++ )
                    _pancakes[i] = x._pancakes[i]->clone();
                //_pancakes     = x._pancakes;
                _cavitydists    = x._cavitydists;
                _beliefs        = x._beliefs;
                _ij             = x._ij;
            }
            return *this;
        }
        ~LCBP2() {
            for( size_t i = 0; i < _pancakes.size(); i++ )
                if( _pancakes[i] != NULL ) {
                    delete _pancakes[i];
                    _pancakes[i] = NULL;
                }
            _pancakes.clear();
        }

        double CalcCavityDist (size_t i, const string &name, const Properties &opts);
        double InitCavityDists (const string &name, const Properties &opts);

        void init();
        void UpdatePancake( size_t ij );
        double run();

        static const char *Name;
        string identify() const;
        Factor belief (const Var &n) const;
        Factor belief (const VarSet &ns) const { assert( 0 == 1 ); }
        vector<Factor> beliefs() const;
        Complex logZ() const { return NAN; }
        void CalcBelief (size_t i);
        const Factor & belief (size_t i) const { return _beliefs[i]; };
//      const BP & pancake (size_t i) const { return _pancakes[i]; };
        const vector<Factor> & cavitydists (size_t i) const { return _cavitydists[i]; };
        size_t nr_ij() const { return _ij.size(); };

        void clamp( const Var &n, size_t i ) { assert( 0 == 1 ); }
        void undoProbs( const VarSet &ns ) { assert( 0 == 1 ); }
        void saveProbs( const VarSet &ns ) { assert( 0 == 1 ); }
        void makeFactorCavity(size_t I) { assert( 0 == 1 ); }
        virtual void makeCavity(const Var & n) { assert( 0 == 1 ); }
        bool checkProperties();
};


#endif
