/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <algorithm>
#include <map>
#include <set>
#include "lcbp3.h"
#include "diffs.h"
#include "util.h"
#include "alldai.h"


using namespace std;


const char *LCBP3::Name = "LCBP3";


bool LCBP3::checkProperties() {
    if( !HasProperty("type") )
        return false;
    if( !HasProperty("cavity") )
        return false;
    if( !HasProperty("updates") )
        return false;
    if( !HasProperty("tol") )
        return false;
    if (!HasProperty("maxiter") )
        return false;
    if (!HasProperty("verbose") )
        return false;

    _properties.type = FromStringTo<TypeType>("type");
    _properties.updates = FromStringTo<UpdateType>("updates");
    _properties.cavity = FromStringTo<CavityType>("cavity");
    ConvertPropertyTo<double>("tol");
    ConvertPropertyTo<size_t>("maxiter");
    ConvertPropertyTo<size_t>("verbose");

    if (HasProperty("cavainame") )
        ConvertPropertyTo<string>("cavainame");
    if (HasProperty("cavaiopts") )
        ConvertPropertyTo<Properties>("cavaiopts");
    if( HasProperty("reinit") )
        _properties.reinit = FromStringTo<bool>("reinit");
    
    return true;
}


LCBP3::LCBP3(const FactorGraph & fg, const Properties &opts) : DAIAlgFG(fg, opts) {
    assert( checkProperties() );

    // create phis
    for( size_t e = 0; e < nrEdges(); e++ )
        _phis.push_back(Factor(var(edge(e).first)));

    // create pancakes
    _pancakes.resize(nrFactors());
   
    // create cavitydists
    for( size_t I=0; I < nrFactors(); I++ )
        _cavitydists.push_back( Factor( factor(I).vars()) );

    // create beliefs
    for( size_t i=0; i < nrVars(); i++ )
        _beliefs.push_back(Factor(var(i)));
}


string LCBP3::identify() const { 
    stringstream result (stringstream::out);
    result << Name << GetProperties();
    return result.str();
}


void LCBP3::CalcBelief (size_t i) {
    if( Verbose() >= 4 )
        for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ ) {
            cout << Name << "::" << __func__ << "(" << i << ") neighbour I=" << *I << " gives " << _pancakes[*I].marginal(var(i)) << endl;
        }
    _beliefs[i].fill(1.0);
    for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ ) {
    //    _beliefs[i] *= _pancakes[*I].marginal(var(i));
        Factor prod;
        for( size_t e = 0; e < nrEdges(); e++ )
            if( edge(e).second == *I )
                prod *= _phis[e];
        _beliefs[i] *= (prod * factor(*I) * _cavitydists[*I]).marginal(var(i));
    }
    _beliefs[i] ^= (1.0 / nbV(i).size() );
    _beliefs[i].normalize( _normtype );
    if( Verbose() >= 4 )
        cout << "LCBP3::CalcBelief(" << i << ") has calculated " << _beliefs[i] << endl;
}


double LCBP3::CalcCavityDist (size_t I, const string &name, const Properties &opts) {
    Factor cavI;
    double maxdiff = 0;

    if( Verbose() >= 2 )
        cout << "Initing cavity " << factor(I) << "(" << factor(I).vars().size() << " vars, " << factor(I).stateSpace() << " states)" << endl;

    if( Cavity() == CavityType::UNIFORM ) {
        cavI = Factor( factor(I).vars() );
    } else if( Cavity() == CavityType::RANDOM ) {
        VarSet vars = factor(I).vars();
        for( VarSet::const_iterator v = vars.begin(); v != vars.end(); v++ )
            cavI *= Factor(*v).randomize();
    } else {
        InfAlg *cav = newInfAlg( name, *this, opts );
        cav->makeFactorCavity( I );
        if( Cavity() == CavityType::FULL )
            cavI = calcMarginal( *cav, factor(I).vars(), reInit() );
        else if( Cavity() == CavityType::PAIR ) {
            cavI = calcMarginal2ndO( *cav, factor(I).vars(), reInit() );
        }
        maxdiff = cav->MaxDiff();
        delete cav;
    }
    cavI.normalize( _normtype );
    _cavitydists[I] = cavI;

    return maxdiff;
}


double LCBP3::InitCavityDists (const string &name, const Properties &opts) {
    double tic = toc();

    if( Verbose() >= 1 ) {
        cout << "LCBP3::InitCavityDists:  ";
        if( Cavity() == CavityType::UNIFORM )
            cout << "Using uniform initial cavity distributions" << endl;
        else if( Cavity() == CavityType::FULL )
            cout << "Using full " << name << opts << "...";
        else if( Cavity() == CavityType::PAIR )
            cout << "Using pairwise " << name << opts << "...";
    }

    double maxdiff = 0.0;
    for( size_t I = 0; I < nrFactors(); I++ ) {
        double md = CalcCavityDist(I, name, opts);
        if( md > maxdiff )
            maxdiff = md;
    }
    init();

    if( Verbose() >= 1 ) {
        cout << "used " << toc() - tic << " clocks." << endl;
    }

    return maxdiff;
}


Factor LCBP3::Pancake (size_t I) {
    Factor result(factor(I));
    result *= _cavitydists[I];
    for( size_t e = 0; e < nrEdges(); e++ )
        if( edge(e).second == I )
            result *= _phis[e];
    return result;
}


void LCBP3::init() {
    if( Updates() == UpdateType::SEQRND )
        for( size_t e = 0; e < nrEdges(); e++ )
            _phis[e].randomize();
    else
        for( size_t e = 0; e < nrEdges(); e++ )
            _phis[e].fill(1.0);

    for( size_t I = 0; I < nrFactors(); I++ )
        _pancakes[I] = Pancake(I).normalized(_normtype);

    for( size_t i = 0; i < nrVars(); i++ )
        CalcBelief(i);
}


Factor LCBP3::NewPancake (size_t iI, bool & hasNaNs) {
    size_t i = edge(iI).first;
    size_t I = edge(iI).second;

    Factor piet = _pancakes[I];

    Factor margi;
    _nb_t nbi = nbV(i);

/*    for( _nb_cit J = nbi.begin(); J != nbi.end(); J++ )
        margi *= _pancakes[*J].marginal( var(i) ).divided_by((_pancakes[*J] * factor(*J).inverse()).marginal(var(i)));
    Factor quot = margi.divided_by( _pancakes[I].marginal( var(i) ) );*/

/*    for( _nb_cit J = nbi.begin(); J != nbi.end(); J++ )
        margi *= _pancakes[*J].marginal( var(i) );
    for( size_t e = 0; e < nrEdges(); e++ )
        if( edge(e).first == i )
            margi /= _phis[e];
    Factor quot = margi.divided_by( _pancakes[I].marginal( var(i) ) );*/

    for( _nb_cit J = nbi.begin(); J != nbi.end(); J++ )
        margi *= _pancakes[*J].marginal( var(i) );
    Factor quot = (margi^(1.0/nbi.size())).divided_by( _pancakes[I].marginal( var(i) ) );
    quot.normalize( _normtype );

/*    for( _nb_cit J = nbi.begin(); J != nbi.end(); J++ )
        if( *J != I )
            margi *= _pancakes[*J].marginal( var(i) );
    margi ^= 1.0 / (nbi.size() - 1);
    Factor quot = margi.divided_by( (_pancakes[I] * factor(I).inverse()).marginal(var(i)) );*/

/*    for( _nb_cit J = nbi.begin(); J != nbi.end(); J++ )
        if( *J != I ) {
            Factor prod;
            for( size_t e = 0; e < nrEdges(); e++ )
                if( (edge(e).second == *J) && (edge(e).first != i) )
                    prod *= _phis[e];
            margi *= (factor(*J) * prod * _cavitydists[*J].marginal(factor(*J).vars() / var(i))).marginal(var(i));
        }
    Factor quot = margi.divided_by(_phis[iI]);*/

    piet *= quot;
    _phis[iI] *= quot;

    piet.normalize( _normtype );
    _phis[iI].normalize( _normtype );

    if( piet.hasNaNs() ) {
        cout << "LCBP3::NewPancake(" << iI << "):  has NaNs!" << endl;
        hasNaNs = true;
    }

    double d = dist( piet.normalized(), Pancake(I).normalized(), Prob::DISTLINF );
//    cout << d << endl;
    assert( d < Tol() );

    return piet;
}


double LCBP3::run() {
    if( Verbose() >= 1 )
        cout << "Starting " << identify() << "...";
    if( Verbose() >= 2 )
        cout << endl;

    double tic = toc();
    Diffs diffs(nrVars(), 1.0);

    updateMaxDiff( InitCavityDists(GetPropertyAs<string>("cavainame"), GetPropertyAs<Properties>("cavaiopts")) );

    vector<Factor> old_beliefs;
    for(size_t i=0; i < nrVars(); i++ )
        old_beliefs.push_back(belief(i));

    bool hasNaNs = false;
    for( size_t I=0; I < nrFactors(); I++ )
        if( _pancakes[I].hasNaNs() ) {
            hasNaNs = true;
            break;
        }
    if( hasNaNs ) {
        cout << Name << "::" << __func__ << ":  initial _pancakes has NaNs!" << endl;
        return NAN;
    }

    vector<long> update_seq(nrEdges(),0);
    for( size_t k=0; k < nrEdges(); k++ )
        update_seq[k] = k;

    size_t iter=0;

    // do several passes over the network until maximum number of iterations has
    // been reached or until the maximum belief difference is smaller than tolerance
    for( iter=0; iter < MaxIter() && diffs.max() > Tol(); iter++ ) {
        // Sequential updates
        if( Updates() == UpdateType::SEQRND )
            random_shuffle( update_seq.begin(), update_seq.end() );
        
        for( size_t t=0; t < nrEdges(); t++ ) {
            long iI = update_seq[t];
            long I = edge(iI).second;
            _pancakes[I] = NewPancake(iI, hasNaNs);
            if( hasNaNs )
                return NAN;
            for( size_t i = 0; i < nrVars(); i++ )
                CalcBelief(i); //FIXME: this wastes cpu time
        }

        // compare new beliefs with old ones
        for(size_t i=0; i < nrVars(); i++ ) {
            diffs.push( dist( belief(i), old_beliefs[i], Prob::DISTLINF ) );
            old_beliefs[i] = belief(i);
        }

        if( Verbose() >= 3 )
            cout << "LCBP3::run:  maxdiff " << diffs.max() << " after " << iter+1 << " passes" << endl;
    }

    Iterations( iter );
    updateMaxDiff( diffs.max() );

    if( Verbose() >= 1 ) {
        if( diffs.max() > Tol() ) {
            if( Verbose() == 1 )
                cout << endl;
                cout << "LCBP3::run:  WARNING: not converged within " << MaxIter() << " passes (" << toc() - tic << " clocks)...final maxdiff:" << diffs.max() << endl;
        } else {
            if( Verbose() >= 2 )
                cout << "LCBP3::run:  ";
                cout << "converged in " << iter << " passes (" << toc() - tic << " clocks)." << endl;
        }
    }

    return diffs.max();
}


Factor LCBP3::belief (const Var &n) const {
    return( _beliefs[findVar(n)] );
}


vector<Factor> LCBP3::beliefs() const {
    return _beliefs;
}
