/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <algorithm>
#include <map>
#include <set>
#include "lcbp4.h"
#include "diffs.h"
#include "util.h"
#include "alldai.h"


using namespace std;


const char *LCBP4::Name = "LCBP4";


bool LCBP4::checkProperties() {
    if( !HasProperty("cavity") )
        return false;
    if( !HasProperty("updates") )
        return false;
    if( !HasProperty("tol") )
        return false;
    if (!HasProperty("maxiter") )
        return false;
    if (!HasProperty("verbose") )
        return false;

    _properties.updates = FromStringTo<UpdateType>("updates");
    _properties.cavity = FromStringTo<CavityType>("cavity");
    ConvertPropertyTo<double>("tol");
    ConvertPropertyTo<size_t>("maxiter");
    ConvertPropertyTo<size_t>("verbose");

    if (HasProperty("cavainame") )
        ConvertPropertyTo<string>("cavainame");
    if (HasProperty("cavaiopts") )
        ConvertPropertyTo<Properties>("cavaiopts");
    if( HasProperty("reinit") )
        _properties.reinit = FromStringTo<bool>("reinit");
    
    return true;
}


LCBP4::LCBP4(const FactorGraph & fg, const Properties &opts) : DAIAlgFG(fg, opts) {
    assert( checkProperties() );

    // Construct _del, _delI and _phis
    _del.resize( nrVars() );
    _delI.resize( nrVars() );
    _phis.resize( nrVars() );
    _nr_ij = 0;
    for( size_t i = 0; i < nrVars(); i++ ) {
        if( Verbose() >= 3 )
            cout << "i = " << i << ", delta(var(i)) = " << delta(var(i)) << endl;
        _del[i].reserve( delta(var(i)).size() );
        _delI[i].reserve( delta(var(i)).size() );
        _phis[i].reserve( delta(var(i)).size() );
        for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ )
            for( _nb_cit j = nbF(*I).begin(); j != nbF(*I).end(); j++ ) 
                if( *j != i ) {
                    _del[i].push_back( *j );
                    _delI[i].push_back( *I );
                    _phis[i].push_back( Factor( var(*j) ) );
                    _nr_ij++;
                }
    }

    // Construct _zeta
    /// _zeta[i][_k][_j] corresponds to \f$ \zeta^{\setminus i}(x_k | x_j) \f$
    _zeta.resize( nrVars() );
    for( size_t i = 0; i < nrVars(); i++ ) {
        _zeta[i].resize( _del[i].size() );
        for( size_t _k = 0; _k < _del[i].size(); _k++ ) {
            _zeta[i][_k].resize( _del[i].size() );
            for( size_t _j = 0; _j < _del[i].size(); _j++ )
                _zeta[i][_k][_j] = Factor( var(_del[i][_k]) | var(_del[i][_j]) );
        }
    }

    // create beliefs
    _beliefs.reserve( nrVars() );
    for( size_t i=0; i < nrVars(); i++ )
        _beliefs.push_back(Factor(var(i)));

    if( Verbose() >= 3 ) {
        cout << "_del: " << endl;
        for( size_t i = 0; i < nrVars(); i++ )
            for( size_t _j = 0; _j < _del[i].size(); _j++ )
                cout << "(" << i << "," << _del[i][_j] << ") ";
        cout << endl;

        cout << "_delI: " << endl;
        for( size_t i = 0; i < nrVars(); i++ )
            for( size_t _j = 0; _j < _delI[i].size(); _j++ )
                cout << "(" << i << "," << factor(_delI[i][_j]).vars() << ") ";
        cout << endl;

        cout << "_phis: " << endl;
        for( size_t i = 0; i < nrVars(); i++ )
            for( size_t _j = 0; _j < _phis[i].size(); _j++ )
                cout << "(" << i << "," << _phis[i][_j].vars() << ") ";
        cout << endl;

        cout << "_nr_ij = " << _nr_ij << endl;
    }
}


string LCBP4::identify() const { 
    stringstream result (stringstream::out);
    result << Name << GetProperties();
    return result.str();
}


void LCBP4::CalcBelief (size_t i) {
    Factor & psi_i = factor(findFactor(var(i)));    // FIXME: such a factor need not exist...

    Factor result(psi_i);

    for( size_t _j = 0; _j < _del[i].size(); _j++ ) {
        size_t j  = _del[i][_j];
        size_t ij = _delI[i][_j];
        Factor part(var(i),1.0);

        for( size_t _k = 0; _k < _del[i].size(); _k++ ) 
            if( _k != _j ) {
                size_t ik = _delI[i][_k];
                part *= (_zeta[i][_k][_j] * factor(ik) * _phis[i][_k]).partSum( var(i) | var(j) );
            }
        part *= _phis[i][_j] * factor( ij );

        result *= part.partSum(var(i)) ^ (1.0 / _del[i].size());
    }

    _beliefs[i] = result.normalized();

/*    Factor correction( var(i), 1.0 );

    for( size_t _k1 = 0; _k1 < _del[i].size(); _k1++ ) {
        size_t ik1 = _delI[i][_k1];
        for( size_t _k2 = _k1 + 1; _k2 < _del[i].size(); _k2++ ) {
            size_t ik2 = _delI[i][_k2];
            correction += (_phis[i][_k1] * _phis[i][_k2] * factor(ik1) * factor(ik2) * cavpair(i,_k1,_k2)).partSum(var(i)).divided_by( F(i,_k1) * F(i,_k2) );
        }
    }

    if( Verbose() >= 3 )
        cout << correction << endl;

//    _beliefs[i] = (psi_i * Fdel(i)).normalized();
    _beliefs[i] = (psi_i * Fdel(i) * correction).normalized();
*/
}


double LCBP4::CalcCavityDist (size_t i, const string &name, const Properties &opts) {
    double maxdiff = 0;

    if( Verbose() >= 2 )
        cout << "Initing cavity " << var(i) << "(" << delta(var(i)).size() << " vars, " << delta(var(i)).stateSpace() << " states)" << endl;

    if( Cavity() == CavityType::UNIFORM ) {
        for( size_t _k = 0; _k < _del[i].size(); _k++ ) {
            for( size_t _j = 0; _j < _del[i].size(); _j++ )
                if( _j != _k )
                    _zeta[i][_k][_j] = Factor( var(_del[i][_k]) | var(_del[i][_j]), 1.0 );
        }
        maxdiff = 0.0;
    } else if( Cavity() == CavityType::PAIR || Cavity() == CavityType::PAIR2 ) {
        InfAlg *cav = newInfAlg( name, *this, opts );
        cav->makeCavity( var(i) );

        vector<Factor> pairbeliefs;
        if( Cavity() == CavityType::PAIR ) 
            pairbeliefs = calcPairBeliefs( *cav, delta(var(i)), reInit() );
        else if( Cavity() == CavityType::PAIR2 )
            pairbeliefs = calcPairBeliefsNew( *cav, cav->delta(var(i)), reInit() );
        maxdiff = cav->MaxDiff();
        delete cav;
        
        // store pairbeliefs in _cavitydists
        VarSet ns = delta(var(i));
        size_t ind = 0;
        VarSet::const_iterator nj = ns.begin();
        for( VarSet::const_iterator nj = ns.begin(); nj != ns.end(); nj++ ) {
            // find _j such that _del[i][_j] == *nj
            size_t _j = 0;
            for( _j = 0; _j < _del[i].size(); _j++ )
                if( var(_del[i][_j]) == *nj )
                    break;
            assert( _j < _del[i].size() );

            for( VarSet::const_iterator nk = nj; (++nk) != ns.end(); ) {
                // find _k such that _del[i][_k] == *nk
                size_t _k = 0;
                for( _k = 0; _k < _del[i].size(); _k++ )
                    if( var(_del[i][_k]) == *nk )
                        break;
                assert( _k < _del[i].size() );
                
                assert( pairbeliefs[ind].vars() == (*nj | *nk) );

                _zeta[i][_j][_k] = pairbeliefs[ind].normalized().divided_by(pairbeliefs[ind].marginal(*nj) * pairbeliefs[ind].marginal(*nk));
                _zeta[i][_k][_j] = _zeta[i][_j][_k];

//                cavpair(i, _j, _k) = RemoveFirstOrderInteractions(pairbeliefs[ind]).log();
//                cavpair(i, _j, _k) -= Factor( *nj | *nk, cavpair(i, _j, _k).totalSum() / (nj->states() * nk->states()) );

                if( Verbose() >= 3 )
                    cout << pairbeliefs[ind] << " wordt " << _zeta[i][_k][_j] << endl;

                ind++;
            }
        }
    } else {
        throw "Not implemented yet";
    }

    return maxdiff;
}


double LCBP4::InitCavityDists (const string &name, const Properties &opts) {
    double tic = toc();

    if( Verbose() >= 1 )
        cout << Name << "::InitCavityDists starting..." << endl;

    double maxdiff = 0.0;
    for( size_t i = 0; i < nrVars(); i++ ) {
        double md = CalcCavityDist(i, name, opts);
        if( md > maxdiff )
            maxdiff = md;
    }
    init();

    if( Verbose() >= 1 )
        cout << Name << "::InitCavityDists:  used " << toc() - tic << " clocks." << endl;

    return maxdiff;
}


void LCBP4::init() {
    // Init _phis
    for( size_t i = 0; i < nrVars(); i++ )
        for( size_t _j = 0; _j < _del[i].size(); _j++ ) {
            Factor & phi = _phis[i][_j];

            if( Updates() == UpdateType::SEQRND )
                phi.randomize();
            else
                phi.fill(1.0);
        }

    // Calc initial beliefs
    for( size_t i = 0; i < nrVars(); i++ )
        CalcBelief(i);
}


void LCBP4::update( size_t i, size_t _j ) {
    // recalculate _phis[i][_j] in cavity i
    
    size_t j = _del[i][_j];
    size_t _i = 0;
    for( _i = 0; _i < _del[j].size(); _i++ )
        if( _del[j][_i] == i )
            break;
    assert( _i != _del[j].size() );
    VarSet ij = var(i) | var(j);

    Factor & psi_i = factor(findFactor(var(i)));
    Factor & psi_j = factor(findFactor(var(j)));

    Factor numer(var(j), 1.0);
    for( size_t _m = 0; _m < _del[j].size(); _m++ )
        if( _m != _i ) {
            size_t jm = _delI[j][_m];
            numer *= (_zeta[j][_m][_i] * factor(jm) * _phis[j][_m]).partSum( ij );
        }
    numer = (numer * _phis[j][_i]).partSum( var(j) );

    Factor denom(var(j), 1.0);
    for( size_t _k = 0; _k < _del[i].size(); _k++ )
        if( _k != _j ) {
            size_t ik = _delI[i][_k];
            denom *= (_zeta[i][_k][_j] * factor(ik) * _phis[i][_k]).partSum( ij );
        }
    denom = (denom * psi_i).partSum( var(j) );

//    Factor result( psi_j );
//    result *= numer;
//    result *= denom.inverse();  // FIXME: why cannot I just use operator/=, or operator/ ?
//    result.normalize();
    _phis[i][_j] = (psi_j * numer / denom).normalized();

/*
    // F(j,_k) Corresponds to \f$ F^k_j(x_j) = \sum_{x_k} \psi_{jk}(x_j,x_k) \Phi^{\setminus j}_k (x_k) \f$

    Factor correction( var(j), 1.0 );

    for( size_t _l = 0; _l < _del[j].size(); _l++ ) 
        if( _l != _i ) {
            size_t jl = _delI[j][_l];
            correction += (_phis[j][_i] * _phis[j][_l] * cavpair(j,_i,_l) * factor(jl)).partSum( var(j) ).divided_by( F(j,_l) ) / _phis[j][_i].totalSum();
        }

    for( size_t _l1 = 0; _l1 < _del[j].size(); _l1++ )
        if( _l1 != _i ) {
            size_t jl1 = _delI[j][_l1];
            for( size_t _l2 = _l1 + 1; _l2 < _del[j].size(); _l2++ )
                if( _l2 != _i ) {
                    size_t jl2 = _delI[j][_l2];
                    correction += (_phis[j][_l1] * _phis[j][_l2] * factor(jl1) * factor(jl2) * cavpair(j,_l1,_l2)).partSum(var(j)).divided_by( F(j,_l1) * F(j,_l2) );
                }
        }

    double denom = (psi_i * Fdele(i,_j)).totalSum();
    for( size_t _m = 0; _m < _del[i].size(); _m++ )
        if( _m != _j ) {
            size_t im = _delI[i][_m];
            correction -= (Fdele(i,_j,_m) * psi_i * (_phis[i][_m] * factor(im) * cavpair(i,_j,_m)).partSum( ij ) ).partSum( var(j) ) / denom;
        }


    for( size_t _m1 = 0; _m1 < _del[i].size(); _m1++ )
        if( _m1 != _j ) {
            size_t im1 = _delI[i][_m1];
            for( size_t _m2 = _m1 + 1; _m2 < _del[i].size(); _m2++ )
                if( _m2 != _j ) {
                    size_t im2 = _delI[i][_m2];
                    double c = (Fdele(i,_j,_m1,_m2) * psi_i * (_phis[i][_m1] * factor(im1) * _phis[i][_m2] * factor(im2) * cavpair(i,_m1,_m2)).partSum(var(i)) ).totalSum() / denom;
                    correction -= Factor( var(j), c );
                }
        }

    if( Verbose() >= 3 )
        cout << correction << endl;

//    _phis[i][_j] = (psi_j * Fdele(j,_i)).normalized();
    _phis[i][_j] = (psi_j * Fdele(j,_i) * correction).normalized();
*/
}


double LCBP4::run() {
    if( Verbose() >= 1 )
        cout << "Starting " << identify() << "...";
    if( Verbose() >= 2 )
        cout << endl;

    double tic = toc();
    Diffs diffs(nrVars(), 1.0);

    updateMaxDiff( InitCavityDists(GetPropertyAs<string>("cavainame"), GetPropertyAs<Properties>("cavaiopts")) );

    vector<Factor> old_beliefs;
    for(size_t i=0; i < nrVars(); i++ )
        old_beliefs.push_back(belief(i));

    typedef pair<size_t,size_t> i_j;
    vector<i_j> update_seq;
    update_seq.reserve( _nr_ij );
    for( size_t i=0; i < nrVars(); i++ )
        for( size_t _j=0; _j < _del[i].size(); _j++ )
            update_seq.push_back( i_j(i,_j) );

    // do several passes over the network until maximum number of iterations has
    // been reached or until the maximum belief difference is smaller than tolerance
    size_t iter=0;
    for( iter=0; iter < MaxIter() && diffs.max() > Tol(); iter++ ) {
        // Sequential updates
        if( Updates() == UpdateType::SEQRND )
            random_shuffle( update_seq.begin(), update_seq.end() );
        
        for( size_t t=0; t < _nr_ij; t++ ) {
            size_t i = update_seq[t].first;
            size_t _j = update_seq[t].second;
            update(i,_j);
        }

        // compare new beliefs with old ones
        for(size_t i=0; i < nrVars(); i++ ) {
            CalcBelief(i);
            diffs.push( dist( belief(i), old_beliefs[i], Prob::DISTLINF ) );
            old_beliefs[i] = belief(i);
        }

        if( Verbose() >= 3 )
            cout << "LCBP4::run:  maxdiff " << diffs.max() << " after " << iter+1 << " passes" << endl;
    }

    Iterations( iter );
    updateMaxDiff( diffs.max() );

    if( Verbose() >= 1 ) {
        if( diffs.max() > Tol() ) {
            if( Verbose() == 1 )
                cout << endl;
                cout << "LCBP4::run:  WARNING: not converged within " << MaxIter() << " passes (" << toc() - tic << " clocks)...final maxdiff:" << diffs.max() << endl;
        } else {
            if( Verbose() >= 2 )
                cout << "LCBP4::run:  ";
                cout << "converged in " << iter << " passes (" << toc() - tic << " clocks)." << endl;
        }
    }

    return diffs.max();
}
