/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands

    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __LCBP4_H__
#define __LCBP4_H__


#include "daialg.h"
#include "enum.h"
#include "factorgraph.h"


using namespace std;


// CURRENTLY ONLY FOR PAIRWISE INTERACTIONS!


class LCBP4 : public DAIAlgFG {
    protected:
        /// _del[i] is a vector with the indices of the neighbouring variables of i
        vector<vector<size_t> >      _del;
        /// _delI[i] is a vector with the indices of the neighbouring factors of i,
        /// corresponding precisely to the neighbouring variables of i, i.e.
        /// factor(_delI[i][ij]) consists of var(i) and var(_del[i][ij])
        vector<vector<size_t> >      _delI;
        /// Total number of ordered pairs (i,j) with j in delta(i)
        size_t _nr_ij;

        /// _phis[i][_j] corresponds to \f$ \Phi^{\setminus i}_j(x_j) \f$
        vector<vector<Factor> >      _phis;
        
        /// _zeta[i][_k][_j] corresponds to \f$ \zeta^{\setminus i}(x_k | x_j) \f$
        vector<vector<vector<Factor > > >   _zeta;

        /// Single variable beliefs
        vector<Factor>      _beliefs;


    public:
        ENUM(CavityType,PAIR,PAIR2,PAIRINT,PAIRCUM,UNIFORM)
        ENUM(UpdateType,SEQFIX,SEQRND,NONE)

        struct {
            UpdateType updates;
            CavityType cavity;
            bool       reinit;
        } _properties;
        CavityType Cavity() const { return _properties.cavity; }
        UpdateType Updates() const { return _properties.updates; }
        bool reInit() const { return _properties.reinit; }
        
        /// Default constructor
        LCBP4() : DAIAlgFG() {};
        /// Copy constructor
        LCBP4(const LCBP4 & x) : DAIAlgFG(x), _del(x._del), _delI(x._delI), _nr_ij(x._nr_ij), _phis(x._phis), _zeta(x._zeta), _beliefs(x._beliefs) {};
        /// Clone function
        LCBP4* clone() const { return new LCBP4(*this); }
        /// Create function
        LCBP4* create() const { return new LCBP4(); }
        /// Construct LCBP4 object from a FactorGraph and parameters
        LCBP4(const FactorGraph & fg, const Properties &opts);
        /// Assignment operator
        LCBP4& operator=(const LCBP4 & x) {
            if( this != &x ) {
                DAIAlgFG::operator=(x);
                _del            = x._del;
                _delI           = x._delI;
                _nr_ij          = x._nr_ij;
                _phis           = x._phis;
                _zeta           = x._zeta;
                _beliefs        = x._beliefs;
            }
            return *this;
        }

        static const char *Name;
        double CalcCavityDist (size_t i, const string &name, const Properties &opts);
        double InitCavityDists (const string &name, const Properties &opts);

        void init();
        double run();
        void update( size_t i, size_t _j );
        /// F(j,_k) Corresponds to \f$ F^k_j(x_j) = \sum_{x_k} \psi_{jk}(x_j,x_k) \Phi^{\setminus j}_k (x_k) \f$
/*        Factor F( size_t j, size_t _k ) { return (factor(_delI[j][_k]) * _phis[j][_k]).partSum( var(j) ); }
        Factor Fdel( size_t j ) { 
            Factor result;
            for( size_t _k = 0; _k < _del[j].size(); _k++ )
                result *= F(j, _k);
            return result;
        }
        Factor Fdele( size_t j, size_t _e1 ) { 
            Factor result;
            for( size_t _k = 0; _k < _del[j].size(); _k++ )
                if( _k != _e1 )
                    result *= F(j, _k);
            return result;
        }
        Factor Fdele( size_t j, size_t _e1, size_t _e2 ) { 
            Factor result;
            for( size_t _k = 0; _k < _del[j].size(); _k++ )
                if( (_k != _e1) && (_k != _e2) )
                    result *= F(j, _k);
            return result;
        }
        Factor Fdele( size_t j, size_t _e1, size_t _e2, size_t _e3 ) { 
            Factor result;
            for( size_t _k = 0; _k < _del[j].size(); _k++ )
                if( (_k != _e1) && (_k != _e2) && (_k != _e3) )
                    result *= F(j, _k);
            return result;
        }
        Factor & cavpair( size_t i, size_t _k1, size_t _k2 ) {
            if( _k1 < _k2 )
                return _zeta[i][_k1][_k2];
            else if( _k1 > _k2 )
                return _zeta[i][_k2][_k1];
            else
                throw "Internal error in LCBP4::cavpair: _k1 == _k2";
        }
*/
        string identify() const;
        Factor belief (const Var &n) const { return( _beliefs[findVar(n)] ); }
        Factor belief (const VarSet &ns) const { assert( 0 == 1 ); }
        vector<Factor> beliefs() const { return _beliefs; }
        Complex logZ() const { return NAN; }
        void CalcBelief (size_t i);
        const Factor & belief (size_t i) const { return _beliefs[i]; };

        void clamp( const Var &n, size_t i ) { assert( 0 == 1 ); }
        void undoProbs( const VarSet &ns ) { assert( 0 == 1 ); }
        void saveProbs( const VarSet &ns ) { assert( 0 == 1 ); }
        void makeFactorCavity(size_t I) { assert( 0 == 1 ); }
        virtual void makeCavity(const Var & n) { assert( 0 == 1 ); }
        bool checkProperties();
};


#endif
