/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <algorithm>
#include <map>
#include <set>
#include "lcbp5.h"
#include "diffs.h"
#include "util.h"
#include "alldai.h"


using namespace std;


const char *LCBP5::Name = "LCBP5";


bool LCBP5::checkProperties() {
    if( !HasProperty("cavity") )
        return false;
    if( !HasProperty("updates") )
        return false;
    if( !HasProperty("tol") )
        return false;
    if (!HasProperty("maxiter") )
        return false;
    if (!HasProperty("verbose") )
        return false;

    _properties.updates = FromStringTo<UpdateType>("updates");
    _properties.cavity = FromStringTo<CavityType>("cavity");
    ConvertPropertyTo<double>("tol");
    ConvertPropertyTo<size_t>("maxiter");
    ConvertPropertyTo<size_t>("verbose");

    if (HasProperty("cavainame") )
        ConvertPropertyTo<string>("cavainame");
    if (HasProperty("cavaiopts") )
        ConvertPropertyTo<Properties>("cavaiopts");
    if( HasProperty("reinit") )
        _properties.reinit = FromStringTo<bool>("reinit");
    
    return true;
}


LCBP5::LCBP5(const FactorGraph & fg, const Properties &opts) : DAIAlgFG(fg, opts) {
    assert( checkProperties() );

    // Construct _del, _delI and _phis
    _del.resize( nrVars() );
    _delI.resize( nrVars() );
    _phis.resize( nrVars() );
    _nr_ij = 0;
    for( size_t i = 0; i < nrVars(); i++ ) {
        _del[i].reserve( delta(var(i)).size() );
        _delI[i].reserve( delta(var(i)).size() );
        _phis[i].reserve( delta(var(i)).size() );
        for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ )
            for( _nb_cit j = nbF(*I).begin(); j != nbF(*I).end(); j++ ) 
                if( *j != i ) {
                    _del[i].push_back( *j );
                    _delI[i].push_back( *I );
                    _phis[i].push_back( Factor( var(*j) ) );
                    _nr_ij++;
                }
    }

    // Construct _gamma
    /// _gamma[i][_k1][_k2] corresponds to \f$ \zeta^{\setminus i}_{_k1,_k2} \f$ for k_1 < k_2
    _gamma.resize( nrVars() );
    for( size_t i = 0; i < nrVars(); i++ ) {
        _gamma[i].resize( _del[i].size() );
        for( size_t _k1 = 0; _k1 < _del[i].size(); _k1++ ) {
            _gamma[i][_k1].resize( _del[i].size() );
            for( size_t _k2 = _k1 + 1; _k2 < _del[i].size(); _k2++ )
                _gamma[i][_k1][_k2] = Factor( var(_del[i][_k1]) | var(_del[i][_k2]) );
        }
    }

    // create beliefs
    _beliefs.reserve( nrVars() );
    for( size_t i=0; i < nrVars(); i++ )
        _beliefs.push_back(Factor(var(i)));
}


string LCBP5::identify() const { 
    stringstream result (stringstream::out);
    result << Name << GetProperties();
    return result.str();
}


void LCBP5::CalcBelief (size_t i) {
    Factor & psi_i = factor(findFactor(var(i)));

    Factor correction( var(i), 1.0 );

    for( size_t _k1 = 0; _k1 < _del[i].size(); _k1++ ) {
        size_t ik1 = _delI[i][_k1];
        for( size_t _k2 = _k1 + 1; _k2 < _del[i].size(); _k2++ ) {
            size_t ik2 = _delI[i][_k2];
            correction += (_phis[i][_k1] * _phis[i][_k2] * factor(ik1) * factor(ik2) * gamma(i,_k1,_k2)).partSum(var(i)).divided_by( F(i,_k1) * F(i,_k2) );
        }
    }

    _beliefs[i] = (psi_i * Fdel(i) * correction).normalized( );
}


double LCBP5::CalcCavityDist (size_t i, const string &name, const Properties &opts) {
    double maxdiff = 0;

    if( Verbose() >= 2 )
        cout << "Initing cavity " << var(i) << "(" << delta(var(i)).size() << " vars, " << delta(var(i)).stateSpace() << " states)" << endl;

    if( Cavity() == CavityType::UNIFORM ) {
        for( size_t _k1 = 0; _k1 < _del[i].size(); _k1++ ) {
            for( size_t _k2 = _k1 + 1; _k2 < _del[i].size(); _k2++ )
                _gamma[i][_k1][_k2] = Factor( var(_del[i][_k1]) | var(_del[i][_k2]), 0.0 );
        }
        maxdiff = 0.0;
   } else if( Cavity() == CavityType::PAIR || Cavity() == CavityType::PAIR2 ) {
        InfAlg *cav = newInfAlg( name, *this, opts );
        cav->makeCavity( var(i) );

        vector<Factor> pairbeliefs;
        if( Cavity() == CavityType::PAIR ) 
            pairbeliefs = calcPairBeliefs( *cav, delta(var(i)), reInit() );
        else if( Cavity() == CavityType::PAIR2 )
            pairbeliefs = calcPairBeliefsNew( *cav, cav->delta(var(i)), reInit() );
        maxdiff = cav->MaxDiff();
        delete cav;
        
        // store pairbeliefs in _cavitydists
        VarSet ns = delta(var(i));
        size_t ind = 0;
        VarSet::const_iterator nj = ns.begin();
        for( VarSet::const_iterator nj = ns.begin(); nj != ns.end(); nj++ ) {
            // find _j such that _del[i][_j] == *nj
            size_t _j = 0;
            for( _j = 0; _j < _del[i].size(); _j++ )
                if( var(_del[i][_j]) == *nj )
                    break;
            assert( _j < _del[i].size() );

            for( VarSet::const_iterator nk = nj; (++nk) != ns.end(); ) {
                // find _k such that _del[i][_k] == *nk
                size_t _k = 0;
                for( _k = 0; _k < _del[i].size(); _k++ )
                    if( var(_del[i][_k]) == *nk )
                        break;
                assert( _k < _del[i].size() );
                
                assert( pairbeliefs[ind].vars() == (*nj | *nk) );

                // FIXME: why doesn't operator/ work here?
                gamma(i,_j,_k) = pairbeliefs[ind].normalized().divided_by( pairbeliefs[ind].marginal(*nj) * pairbeliefs[ind].marginal(*nk) );
                gamma(i,_j,_k) -= 1.0;
//                gamma(i, _j, _k) = RemoveFirstOrderInteractions(pairbeliefs[ind]).log();
//                gamma(i, _j, _k) -= Factor( *nj | *nk, gamma(i, _j, _k).totalSum() / (nj->states() * nk->states()) );

                if( Verbose() >= 3 )
                    cout << pairbeliefs[ind] << " wordt " << gamma(i, _j, _k) << endl;

                ind++;
            }
        }
    } else {
        throw "Not implemented yet";

/*        InfAlg *cav = newInfAlg( name, *this, opts );
        cav->makeCavity( var(i) );
        Bi = calcMarginal( *cav, cav->delta(var(i)), reInit() ).normalized();
        maxdiff = cav->MaxDiff();
        delete cav;

        if( Cavity() == CavityType::PAIRINT ) {
            // Set interactions of order > 2 to zero
            size_t N = delta(var(i)).size();
            double *p = &(*Bi.p().p().begin());
            x2x::p2logp (N, p);
            x2x::logp2w (N, p);
            x2x::fill (N, p, 2, 0.0);
            x2x::w2logp (N, p);
//            x2x::logpnorm (N, p);
            x2x::logp2p (N, p);
        } else if( Cavity() == CavityType::PAIRCUM ) {
            // Set cumulants of order > 2 to zero
            size_t N = delta(var(i)).size();
            double *p = &(*Bi.p().p().begin());
            x2x::p2m (N, p);
            x2x::m2c (N, p, N);
            x2x::fill (N, p, 2, 0.0);
            x2x::c2m (N, p, N);
            x2x::m2p (N, p);
        }

        vector<Factor> pairbeliefs;
        if( Cavity() == CavityType::PAIR ) 
            pairbeliefs = calcPairBeliefs( *cav, delta(var(i)), reInit() );
        else if( Cavity() == CavityType::PAIR2 )
            pairbeliefs = calcPairBeliefsNew( *cav, cav->delta(var(i)), reInit() );
        
        // store pairbeliefs in _cavitydists
        VarSet ns = delta(var(i));
        size_t ind = 0;
        VarSet::const_iterator nj = ns.begin();
        for( VarSet::const_iterator nj = ns.begin(); nj != ns.end(); nj++ ) {
            // find _j such that _del[i][_j] == *nj
            size_t _j = 0;
            for( _j = 0; _j < _del[i].size(); _j++ )
                if( var(_del[i][_j]) == *nj )
                    break;
            assert( _j < _del[i].size() );

            for( VarSet::const_iterator nk = nj; (++nk) != ns.end(); ) {
                // find _k such that _del[i][_k] == *nk
                size_t _k = 0;
                for( _k = 0; _k < _del[i].size(); _k++ )
                    if( var(_del[i][_k]) == *nk )
                        break;
                assert( _k < _del[i].size() );
                
                assert( pairbeliefs[ind].vars() == (*nj | *nk) );
                gamma(i, _j, _k) = RemoveFirstOrderInteractions(pairbeliefs[ind]).log();
                gamma(i, _j, _k) -= Factor( *nj | *nk, gamma(i, _j, _k).totalSum() / (nj->states() * nk->states()) );

                if( Verbose() >= 3 )
                    cout << pairbeliefs[ind] << " wordt " << gamma(i, _j, _k) << endl;

                ind++;
            }
        }*/
    }

    return maxdiff;
}


double LCBP5::InitCavityDists (const string &name, const Properties &opts) {
    double tic = toc();

    if( Verbose() >= 1 )
        cout << Name << "::InitCavityDists starting..." << endl;

    double maxdiff = 0.0;
    for( size_t i = 0; i < nrVars(); i++ ) {
        double md = CalcCavityDist(i, name, opts);
        if( md > maxdiff )
            maxdiff = md;
    }
    init();

    if( Verbose() >= 1 )
        cout << Name << "::InitCavityDists:  used " << toc() - tic << " clocks." << endl;

    return maxdiff;
}


void LCBP5::init() {
    // Init _phis
    for( size_t i = 0; i < nrVars(); i++ )
        for( size_t _j = 0; _j < _del[i].size(); _j++ ) {
            Factor & phi = _phis[i][_j];

            if( Updates() == UpdateType::SEQRND )
                phi.randomize();
            else
                phi.fill(1.0);
        }

    // Calc initial beliefs
    for( size_t i = 0; i < nrVars(); i++ )
        CalcBelief(i);
}


void LCBP5::update( size_t i, size_t _j ) {
    // recalculate _phis[i][_j] in cavity i
    
    // F(j,_k) Corresponds to \f$ F^k_j(x_j) = \sum_{x_k} \psi_{jk}(x_j,x_k) \Phi^{\setminus j}_k (x_k) \f$

    size_t j = _del[i][_j];
    size_t _i = 0;
    for( _i = 0; _i < _del[j].size(); _i++ )
        if( _del[j][_i] == i )
            break;
    assert( _i != _del[j].size() );
    VarSet ij = var(i) | var(j);

    Factor & psi_i = factor(findFactor(var(i)));
    Factor & psi_j = factor(findFactor(var(j)));
    Factor & psi_ij = factor(_delI[i][_j]);

    Factor correction( ij, 1.0 );

    for( size_t _k = 0; _k < _del[i].size(); _k++ ) 
        if( _k != _j ) {
            size_t ik = _delI[i][_k];
            correction += (_phis[i][_k] * gamma(i,_j,_k) * factor(ik)).partSum( ij ) * ( F(i,_k).inverse() );
        }

    for( size_t _k = 0; _k < _del[i].size(); _k++ )
        if( _k != _j ) {
            size_t ik = _delI[i][_k];
            for( size_t _l = _k + 1; _l < _del[i].size(); _l++ )
                if( _l != _j ) {
                    size_t il = _delI[i][_l];
                    correction += (gamma(i,_k,_l) * factor(ik) * factor(il) * _phis[i][_k] * _phis[i][_l]).partSum( var(i) ) * ( (F(i,_k) * F(i,_l)).inverse() ) * Factor(var(j),1.0);
                }
        }

    CalcBelief(j);
    _phis[i][_j] = belief(j).divided_by( (psi_i * psi_ij * Fdele(i, _j) * correction).partSum(var(j)) );
}


double LCBP5::run() {
    if( Verbose() >= 1 )
        cout << "Starting " << identify() << "...";
    if( Verbose() >= 2 )
        cout << endl;

    double tic = toc();
    Diffs diffs(nrVars(), 1.0);

    updateMaxDiff( InitCavityDists(GetPropertyAs<string>("cavainame"), GetPropertyAs<Properties>("cavaiopts")) );

    vector<Factor> old_beliefs;
    for(size_t i=0; i < nrVars(); i++ )
        old_beliefs.push_back(belief(i));

    typedef pair<size_t,size_t> i_j;
    vector<i_j> update_seq;
    update_seq.reserve( _nr_ij );
    for( size_t i=0; i < nrVars(); i++ )
        for( size_t _j=0; _j < _del[i].size(); _j++ )
            update_seq.push_back( i_j(i,_j) );

    size_t iter=0;

    // do several passes over the network until maximum number of iterations has
    // been reached or until the maximum belief difference is smaller than tolerance
    for( iter=0; iter < MaxIter() && diffs.max() > Tol(); iter++ ) {
        // Sequential updates
        if( Updates() == UpdateType::SEQRND )
            random_shuffle( update_seq.begin(), update_seq.end() );
        
        for( size_t t=0; t < _nr_ij; t++ ) {
            size_t i = update_seq[t].first;
            size_t _j = update_seq[t].second;
            update(i,_j);
        }

        // compare new beliefs with old ones
        for(size_t i=0; i < nrVars(); i++ ) {
            CalcBelief(i);
            diffs.push( dist( belief(i), old_beliefs[i], Prob::DISTLINF ) );
            old_beliefs[i] = belief(i);
        }

        if( Verbose() >= 3 )
            cout << "LCBP5::run:  maxdiff " << diffs.max() << " after " << iter+1 << " passes" << endl;
    }

    Iterations( iter );
    updateMaxDiff( diffs.max() );

    if( Verbose() >= 1 ) {
        if( diffs.max() > Tol() ) {
            if( Verbose() == 1 )
                cout << endl;
                cout << "LCBP5::run:  WARNING: not converged within " << MaxIter() << " passes (" << toc() - tic << " clocks)...final maxdiff:" << diffs.max() << endl;
        } else {
            if( Verbose() >= 2 )
                cout << "LCBP5::run:  ";
                cout << "converged in " << iter << " passes (" << toc() - tic << " clocks)." << endl;
        }
    }

    return diffs.max();
}
