/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include "bounds.h"
#include "jtree.h"


using namespace std;
using namespace libDAI;


double min( vector<double> & x ) {
    return *min_element( x.begin(), x.end() );
}
double max( vector<double> & x ) {
    return *max_element( x.begin(), x.end() );
}
double avg( vector<double> & x ) {
    double sum = 0.0;
    for( size_t i = 0; i < x.size(); i++ )
        sum += x[i];
    return sum / x.size();
}
double stddev( vector<double> & x ) {
    double sumsq = 0.0;
    for( size_t i = 0; i < x.size(); i++ )
        sumsq += x[i] * x[i];
    double _avg = avg( x ); 
    double result = sumsq / x.size() - _avg * _avg;
    if( result < 0.0 )
        return 0.0;
    else
        return sqrt( result );
}

double vol( const Box &b ) {
    double v = 1.0;
    for( size_t i = 0; i < b.max().size(); i++ )
        v *= (b.max()[i] - b.min()[i]);
    return v;
}
double Dnorm( const Prob &a ) {
    return 0.5 * log( a.maxVal() / a.minVal() );
}
double D( const Prob &a, const Prob &b ) {
    return 0.5 * log( (a / b).maxVal() / (a / b).minVal() );
}
Prob center( const Box &b ) {
    return (b.max() * b.min())^0.5;
}
double diam( const Box &b ) {
    Prob c;
    c = center(b);
    vector<Prob> ep = b.extremePoints();
    double max_D = 0.0;
    for( size_t _ep = 0; _ep < ep.size(); _ep++ ) {
        double D_ep = D( ep[_ep], c );
        if( D_ep > max_D )
            max_D = D_ep;
    }
    max_D *= 2.0;
//    return max_D;
/*  EQUIVALENT, BUT SLOWER:*/
    double max_D2 = 0.0;
    for( size_t _ep1 = 0; _ep1 < ep.size(); _ep1++ ) 
        for( size_t _ep2 = 0; _ep2 < ep.size(); _ep2++ ) { 
            double D_ep1_ep2 = D( ep[_ep1], ep[_ep2] );
            if( D_ep1_ep2 > max_D2 )
                max_D2 = D_ep1_ep2;
        }
//    double max_D3 = D( b.max(), b.min() );  SOMETHING DIFFERENT
    assert( ((1.0-1e-9) * max_D <= max_D2) && (max_D2 <= (1.0+1e-9) * max_D) );
    return max_D;
}
double meas( const Box &b ) {
    Prob c( b.min().size(), 0.0 );
    vector<Prob> ep = b.extremePoints();
    for( size_t _ep = 0; _ep < ep.size(); _ep++ )
        c += ep[_ep].log();
    c /= ep.size();
    c = c.exp();
    double max_D = 0.0;
    for( size_t _ep = 0; _ep < ep.size(); _ep++ ) {
        double D_ep = D( ep[_ep], c );
        if( D_ep > max_D )
            max_D = D_ep;
    }
    return max_D;
}
double diam( const vector<Prob> eps ) {
    double d = 0.0;
    for( size_t ep1 = 0; ep1 < eps.size(); ep1++ )
        for( size_t ep2 = 0; ep2 < eps.size(); ep2++ ) {
            double D_ep1_ep2 = D( eps[ep1], eps[ep2] );
            if( D_ep1_ep2 > d )
                d = D_ep1_ep2;
        }
    return d;
}
double altdiam( const Box &b ) {
/*  vector<Prob> b_ep = b.extremePoints();
    Prob c = center(b);
    double dia = diam( b_ep );
    cout << b.max() << endl;
    cout << b.min() << endl;
    cout << b.max() / b.min() << endl;
    cout << ((b.max() / b.min())^0.5) << endl;
    for( size_t i = 0; i < b_ep.size(); i++ )
        cout << ((fabs(D(b_ep[i],c) - 0.5*dia) < 1e-9) ? "*" : " ") << b_ep[i] << " " << 2.0*D(b_ep[i],c) << endl;
    for( size_t i = 0; i < b_ep.size(); i++ )
        for( size_t j = i+1; j < b_ep.size(); j++ ) {
            if( fabs(D(b_ep[i],b_ep[j]) - dia) < 1e-9 )
                cout << "(" << i << "," << j << ") " << (b_ep[i] / b_ep[j]) << endl;
    }
    */
    Prob q = b.max() / b.min();
    double q1 = q.maxVal();
    double q2 = 1.0;
    for( size_t i = 0; i < q.size(); i++ ) {
        if( q[i] > q2 && q[i] < q1 )
            q2 = q[i];
    } 
    return 0.5 * log(q1 * q2);
}

int main( int argc, char *argv[] ) {
    if( argc != 4 ) {
        cout << "Usage: " << argv[0] << " <in.fg> <varind> <depth>" << endl << endl;
        cout << "Converts a .fg (FactorGraph) file to a .dot (GraphViz) file for visualization," << endl;
        cout << "making a Self Avoiding Walk tree for the variable with index <varind>, up to depth <depth>." << endl;
        cout << "The .dot file can be converted to .ps (PostScript) by 'neato -T ps out.dot > out.ps'" << endl;
        return 1;
    } else {
        for( size_t dim_in = 4; dim_in < 5; dim_in++ )
            for( size_t dim_out = 4; dim_out < 5; dim_out++ ) {
                Var x_in( 0, dim_in );
                Var x_out( 1, dim_out );
                Factor psi( x_in | x_out );
                for( size_t nr = 0; nr < 10000; nr++ ) {
                    psi.randomize();
                    double Dpsi = Dnorm(psi.p());
                    Box b_in( x_in );
                    b_in.min().randomize();
                    b_in.max().randomize();
                    b_in.max() += b_in.min();
                    b_in.normalize();
                    Box b_out = b_in.boundSumProd( psi, x_out, true );

                    vector<Prob> b_in_ep = b_in.extremePoints();
                    vector<Prob> b_out_ep = b_out.extremePoints();

                    Prob b_out_center = center(b_out);
                    Prob b_in_center = center(b_in);
                    double r = 0.5 * diam( b_in );
                    Box ball_in( x_in, b_in_center * exp(-r), b_in_center * exp(r) );
//                    ball_in.normalize();
//                    cout << b_in << " (diam: " << diam(b_in) << ") is subsumed by " << ball_in.normalized() << " (diam: " << diam(ball_in) << ")" << endl;
                    assert( ball_in.normalized().contains( Factor(x_in,b_in_center), 1e-9 ) );
                    for( size_t ep = 0; ep < b_in_ep.size(); ep++ )
                        assert( D( b_in_ep[ep], b_in_center ) <= r );
                    vector<Prob> ball_in_ep = ball_in.extremePoints();
                    double R = 0.0;
                    for( size_t ep = 0; ep < ball_in_ep.size(); ep++ ) {
                        double R_ep = D( (psi * Factor(x_in,ball_in_ep[ep])).marginal( x_out ).p(), (psi * Factor(x_in,b_in_center)).marginal( x_out ).p() );
                        if( R_ep > R )
                            R = R_ep;
                    }
                    assert( R <= r );
                    double S = 0.0;
                    for( size_t ep = 0; ep < ball_in_ep.size(); ep++ ) {
                        double S_ep = D( (psi * Factor(x_in,ball_in_ep[ep])).marginal( x_out ).p(), b_out_center );
                        if( S_ep > S )
                            S = S_ep;
                    }
                    double T = 0.0;
                    for( size_t ep = 0; ep < b_in_ep.size(); ep++ ) {
                        assert( ball_in.normalized().contains( Factor(x_in,b_in_ep[ep]).normalized(), 1e-9 ) );
                        double T_ep = D( (psi * Factor(x_in,b_in_ep[ep])).marginal( x_out ).p(), b_out_center );
                        if( T_ep > T )
                            T = T_ep;
                    }
                    double t = 0.0;
                    for( size_t ep = 0; ep < b_in_ep.size(); ep++ ) {
                        double t_ep = D( b_in_ep[ep], b_in_center );
                        if( t_ep > t )
                            t = t_ep;
                    }
                    double Npsi = psi.strength( x_in, x_out );

                    double d_out, d_in;
                    d_out = diam( b_out );
                    d_in = diam( b_in );
                    vector<Prob> ep_in = b_in_ep;
                    vector<Prob> ep_out;
                    for( size_t i = 0; i < ep_in.size(); i++ )
                        ep_out.push_back( (psi * Factor(x_in,ep_in[i])).marginal(x_out).p() );
                    cout << diam( ep_in ) << "\t" << diam( ep_out ) << "\t" << diam( b_out_ep ) << "\t" << altdiam( b_in ) << endl;
/*                    cout << d_in << " " << d_out << "   " << r << " " << R << " " << S << " " << T << "  " << t << "     " << meas(b_in) << " " << meas(b_out) << endl;
                    assert( t == r );
                    assert( T <= t );*/
//                    assert( S <= R ); not always true
//                    assert( T <= R ); not always true
//                    assert( d_out <= 2.0*R ); not always true

//                  assert( d_out <= (2 * d_in) );

/*                    double max_D_in = 0.0;
                    for( size_t ep = 0; ep < b_in_ep.size(); ep++ ) {
                        double D_ep = D( b_in_ep[ep], b_in_center );
                        if( D_ep > max_D_in )
                            max_D_in = D_ep;
                    }
                    double max_D_out = 0.0;
                    for( size_t ep = 0; ep < b_out_ep.size(); ep++ ) {
                        double D_ep = D( b_out_ep[ep], b_out_center );
                        if( D_ep > max_D_out )
                            max_D_out = D_ep;
                    }
//                  assert( max_D_in >= max_D_out );

                    double max_E_in = 0.0;
                    for( size_t ep = 0; ep < b_in_ep.size(); ep++ )
                        for( size_t ep2 = 0; ep2 < b_in_ep.size(); ep2++ ) {
                            double E_ep = D( b_in_ep[ep], b_in_ep[ep2] );
                            if( E_ep > max_E_in )
                                max_E_in = E_ep;
                        }
                    double max_E_out = 0.0;
                    for( size_t ep = 0; ep < b_out_ep.size(); ep++ ) 
                        for( size_t ep2 = 0; ep2 < b_out_ep.size(); ep2++ ) {
                            double E_ep = D( b_out_ep[ep], b_out_ep[ep2] );
                            if( E_ep > max_E_out )
                                max_E_out = E_ep;
                        }

                    cout << Npsi << "\t" << r << "\t" << R << "\t" << log( (exp(2.0*Dpsi) * exp(r) + 1.0) / (exp(2.0*Dpsi) + exp(r)) ) << "\t" << 
                        log( (exp(2.0*atanh(Npsi)) * exp(r) + 1.0) / (exp(2.0*atanh(Npsi)) + exp(r)) ) << "\t" <<
                        0.5 * log( (exp(atanh(Npsi)) * exp(r) + exp(-atanh(Npsi)) * exp(-r)) / (exp(-atanh(Npsi)) * exp(r) + exp(atanh(Npsi)) * exp(-r)) ) << "\t" <<
                        diam( b_in ) << "\t" << diam( b_out ) << "\t" <<
                        max_D_in << "\t" << max_D_out << "\t" <<
                        max_E_in << "\t" << max_E_out << endl;
*/

//                    cout << d_out << " <= " << (2*d_in) << endl;
//                    assert( b_out.gap() <= b_in.gap() );
//                    assert( vol(b_out) <= vol(b_in) );
//                  cout << endl;
                }
            }
        return 0;

        // Read factorgraph
        FactorGraph fg;
        char *infile = argv[1];
        fg.ReadFromFile( infile );

        long varind = atoi( argv[2] );
        size_t depth = atoi( argv[3] );
        double t;
        Box Mooij;
        Box Ihler;

        bool do_exact = false;
        bool do_Ihler = true;
        JTree jt;
        if( varind == -1 ) {
            if( do_exact ) {
                jt = JTree(fg, Properties()("updates",string("HUGIN"))("verbose",string("0UL")));
                jt.init();
                jt.run();
            }
            t = toc();
            vector<double> gaps_Mooij, gaps_Ihler;
            for( size_t i = 0; i < fg.nrVars(); i++ ) {
                Var v_i = fg.var( i );
                BoundTree tre;
                tre.buildTree( fg, v_i, depth*2, true );
                Mooij = tre.BoxProp( fg, false, 0 );
                if( do_Ihler )
                    Ihler = tre.BoxProp( fg, true, 0 );

                /*
                cout << i << ": ";
//                  cout << "Bounding box: " << Mooij << ", ";
                cout << "gap: " << dist( Mooij.max(), Mooij.min(), Prob::DISTLINF );
//                  cout << "exact marginal: " << jt.belief( v_i ) << endl;
                cout << endl;*/

                gaps_Mooij.push_back( dist( Mooij.max(), Mooij.min(), Prob::DISTLINF ) );
                if( do_Ihler )
                    gaps_Ihler.push_back( dist( Ihler.max(), Ihler.min(), Prob::DISTLINF ) );

                if( do_exact ) 
                    assert( Mooij.contains( jt.belief( v_i ), 1e-9 ) );
            }

            cout << min(gaps_Mooij) << " " << max(gaps_Mooij) << " " << avg(gaps_Mooij) << " " << stddev(gaps_Mooij) << " " << toc() - t << "        ";
            if( do_Ihler )
                cout << min(gaps_Ihler) << " " << max(gaps_Ihler) << " " << avg(gaps_Ihler) << " " << stddev(gaps_Ihler) << " " << toc() - t << "        ";
        } else {
            Var v_i = fg.var( varind );
            BoundTree tre;

            tre.buildTree( fg, v_i, depth*2, true );
            tre.display( cout, fg );
            Mooij = tre.BoxProp( fg, true, 0 );

            cerr << "Bounding box: " << Mooij << ", gap: " << dist( Mooij.max(), Mooij.min(), Prob::DISTLINF ) << endl;

            if( do_exact ) {
                jt = JTree(fg, Properties()("updates",string("HUGIN"))("verbose",string("0UL")));
                jt.init();
                jt.run();
                cerr << "Exact marginal: " << jt.belief( v_i ) << endl;

                assert( Mooij.contains( jt.belief( v_i ), 1e-9 ) );
            }
        }

        return 0;
    }
}
