/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*=================================================================*
 *                                                                 * 
 * This is a MEX-file for MATLAB.                                  *
 *                                                                 * 
 *   [A] = dai_bpconvbound(psi,iter,verbose);                      *
 *                                                                 * 
 *=================================================================*/


#include <iostream>
#include "mex.h"
#include "matlab.h"
#include "factorgraph.h"
#include "boxbp.h"


using namespace std;
using namespace libDAI;


/* Input Arguments */

#define PSI_IN          prhs[0]
#define ITER_IN         prhs[1]
#define VERBOSE_IN      prhs[2]
#define NR_IN           3


/* Output Arguments */

#define A_OUT           plhs[0]
#define NR_OUT          1


void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] ) { 
    size_t verbose;
    size_t iter;

    
    // Check for proper number of arguments
    if ((nrhs != NR_IN) || (nlhs != NR_OUT)) { 
        mexErrMsgTxt("Usage: [A] = dai_bpconvbound(psi,iter,verbose);\n\n"
        "\n"
        "INPUT:  psi        = linear cell array containing the factors\n"
        "                     (psi{i} should be a structure with a Member field\n"
        "                     and a P field, like a CPTAB).\n"
        "        iter       = number of iterations of BoxBP.\n"
        "        verbose    = verbosity (0 is quiet, 3 is maximal).\n"
        "\n"
        "OUTPUT: A          = contraction matrix.\n");
    } 
    

    // Get input parameters
    iter = (size_t)*mxGetPr(ITER_IN);
    verbose = (size_t)*mxGetPr(VERBOSE_IN);
    
    vector<Factor> factors = mx2Factors(PSI_IN, verbose);


    // Construct factorgraph
    FactorGraph fg(factors);
    long nr_v = fg.nrVars();
    long nr_f = fg.nrFactors();
    long N = fg.nrEdges();
    typedef vector<FactorGraph::edge_type>::const_iterator edge_cit;
    typedef FactorGraph::nb_cit nb_cit;


    // Count nonzeros
    long nzmax = 0;
    for( edge_cit iI = fg.edges().begin(); iI != fg.edges().end(); iI++ ) {
        size_t i = iI->first;
        size_t I = iI->second;
        for( nb_cit j = fg.nbF(I).begin(); j != fg.nbF(I).end(); j++ )
            if( *j != i )
                for( nb_cit J = fg.nbV(*j).begin(); J != fg.nbV(*j).end(); J++ )
                    if( *J != I )
                        nzmax++;
    }


    // Create A
    mwIndex *Air;
    mwIndex *Ajc;
    double *Apr;
    
    A_OUT = mxCreateSparse(N,N,nzmax,mxREAL);
    Air = mxGetIr(A_OUT);
    Ajc = mxGetJc(A_OUT);
    Apr = mxGetPr(A_OUT);


    // Run BoxProp
    BoxBP boxbp;
    boxbp.run( fg, iter );
    for( size_t iI = 0; iI < fg.nrEdges(); iI++ ) {
        cout << "I = " << fg.edge(iI).second << ", i = " << fg.edge(iI).first << ", box = " << boxbp._boxes[iI] << endl;
    }


    // Init A
    Ajc[0] = 0;
    size_t jc_count = 1;
    size_t count = 0;
    for( size_t iI = 0; iI < N; iI++ ) {
        size_t i = fg.edge(iI).first;
        size_t I = fg.edge(iI).second;
        for( nb_cit j = fg.nbF(I).begin(); j != fg.nbF(I).end(); j++ ) {
            vector<Box> incoming_boxes;
            if( iter != 0 )
                for( nb_cit k = fg.nbF(I).begin(); k != fg.nbF(I).end(); k++ ) 
                    if( *k != i ) {
                        cout << " hIi_box *= boxmsg(" << I << "->" << *k << endl;
                        incoming_boxes.push_back( boxbp._boxes[fg.edge(*k,I)] );
                    }
            Box hIi_box = boundProd( incoming_boxes );
            cout << "I = " << I << ", i = " << i << ", hIi_box = " << hIi_box << endl;

            if( *j != i ) {
                double s = 0.0;
                if( fg.nbV(*j).size() > 1 ) {
                    if( iter == 0 )
                        s = fg.factor(I).strength(fg.var(i), fg.var(*j));
                    else {
                        s = FactorStrength( fg.factor(I), fg.var(i), fg.var(*j), hIi_box );
                    }
                }
                for( nb_cit J = fg.nbV(*j).begin(); J != fg.nbV(*j).end(); J++ )
                    if( *J != I ) {
                        Apr[count] = s;
                        Air[count++] = fg.edge(*j, *J);
                    }
            }
        }
        Ajc[jc_count++] = count;
    }


    return;
}
