/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


/*=================================================================*
 *                                                                 * 
 * This is a MEX-file for MATLAB.                                  *
 *                                                                 * 
 *   N = dai_potstrength(psi,i,j[,hmin,hmax]);                     *
 *                                                                 * 
 *=================================================================*/


#include <iostream>
#include "mex.h"
#include "matlab.h"
#include "factor.h"
#include "box.h"


using namespace std;
using namespace libDAI;


/* Input Arguments */

#define PSI_IN          prhs[0]
#define I_IN            prhs[1]
#define J_IN            prhs[2]
#define HMIN_IN         prhs[3]
#define HMAX_IN         prhs[4]
#define NR_IN           3


/* Output Arguments */

#define N_OUT           plhs[0]
#define NR_OUT          1


void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[] )
{ 
    long ilabel, jlabel;

    // Check for proper number of arguments
    if( ((nrhs != NR_IN) && (nrhs != NR_IN+2)) || (nlhs != NR_OUT) ) { 
        mexErrMsgTxt("Usage: N = dai_potstrength(psi,i,j[,hmin,hmax]);\n\n"
        "\n"
        "INPUT:    psi        = structure with a Member field and a P field, like a CPTAB.\n"
        "          i          = label of a variable in psi.\n"
        "          j          = label of another variable in psi.\n"
        "OPTIONAL: hmin       = lower bound on incoming cavity field (should be factor).\n"
        "          hmax       = upper bound on incoming cavity field (should be factor).\n"
        "\n"
        "OUTPUT:   N          = strength of psi in direction i->j.\n");
    } 
    
    // Get input parameters
    Factor psi = mx2Factor(PSI_IN);
    ilabel = (long)*mxGetPr(I_IN);
    jlabel = (long)*mxGetPr(J_IN);

    // Find variable in psi with label ilabel
    Var i;
    for( VarSet::const_iterator n = psi.vars().begin(); n != psi.vars().end(); n++ )
        if( n->label() == ilabel ) {
            i = *n;
            break;
        }
    assert( i.label() == ilabel );

    // Find variable in psi with label jlabel
    Var j;
    for( VarSet::const_iterator n = psi.vars().begin(); n != psi.vars().end(); n++ )
        if( n->label() == jlabel ) {
            j = *n;
            break;
        }
    assert( j.label() == jlabel );

    // Calculate N(psi,i,j);
    double N;
    if( nrhs == NR_IN )
        N = psi.strength( i, j );
    else {
        Factor hmin = mx2Factor(HMIN_IN);
        Factor hmax = mx2Factor(HMAX_IN);
        assert( psi.vars() / i == hmin.vars() );
        assert( psi.vars() / i == hmax.vars() );
        N = FactorStrength( psi, i, j, Box( psi.vars() / i, hmin.p(), hmax.p() ) );
    }
    
    // Hand over result to MATLAB
    N_OUT = mxCreateDoubleMatrix(1,1,mxREAL);
    *(mxGetPr(N_OUT)) = N;
    
    return;
}
