/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <stdlib.h>
#include <algorithm>
#include "alldai.h"
#include "box.h"
#include "diffs.h"
#include "cmath"
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>
#include "bounds_other.h"

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_multimin.h>


namespace libDAI {

    
    using namespace std;
    using namespace libDAI;


    BinaryPairwiseGM::BinaryPairwiseGM( const FactorGraph &fg ) {
        assert( fg.isPairwise() );
        assert( fg.isBinary() );

        // create w and th
        N = fg.nrVars();

        // count non_zeros in w
        size_t non_zeros = 0;
        for( size_t I = 0; I < fg.nrFactors(); I++ )
            if( fg.factor(I).vars().size() == 2 )
                non_zeros++;
        w = matrix(N, N, non_zeros * 2);

        th = ublasvector(N);
        for( size_t i = 0; i < N; i++ )
            th[i] = 0.0;
        
        logZ0 = 0.0;

        for( size_t I = 0; I < fg.nrFactors(); I++ ) {
            const Factor &psi = fg.factor(I);
            if( psi.vars().size() == 0 )
                logZ0 += log( psi[0] );
            else if( psi.vars().size() == 1 ) {
                size_t i = fg.findVar( *(psi.vars().begin()) );
                th[i] += 0.5 * (log(psi[1]) - log(psi[0]));
                logZ0 += 0.5 * (log(psi[0]) + log(psi[1]));
            } else if( psi.vars().size() == 2 ) {
                size_t i = fg.findVar( *(psi.vars().begin()) );
                VarSet::const_iterator jit = psi.vars().begin();
                size_t j = fg.findVar( *(++jit) );

                double w_ij = 0.25 * (log(psi[3]) + log(psi[0]) - log(psi[2]) - log(psi[1]));
                w(i,j) += w_ij; 
                w(j,i) += w_ij; 

                th[i] += 0.25 * (log(psi[3]) - log(psi[2]) + log(psi[1]) - log(psi[0]));
                th[j] += 0.25 * (log(psi[3]) - log(psi[1]) + log(psi[2]) - log(psi[0]));

                logZ0 += 0.25 * (log(psi[0]) + log(psi[1]) + log(psi[2]) + log(psi[3]));
            }
        }
    }


    double BinaryPairwiseGM::doBP( size_t maxiter, double tol, size_t verbose, ublasvector &m ) {
        double tic = toc();

        if( verbose >= 1 )
            cout << "Starting BinaryPairwiseGM::doBP..." << endl;

        size_t nr_messages = w.nnz();
        matrix message( w );
        for( size_t ij = 0; ij < nr_messages; ij++ )
            message.value_data()[ij] = 0.0;
        // NOTE: message(i,j) is \mu_{j\to i}
        Diffs diffs(nr_messages, 1.0);

        size_t _iterations = 0;
        for( _iterations = 0; _iterations < maxiter && diffs.max() > tol; _iterations++ ) {
            // walk through the sparse array structure
            // this is similar to matlab sparse arrays
            // index2 gives the column index (ir in matlab)
            // index1 gives the starting indices for each row (jc in matlab)
            size_t i = 0;
            for( size_t pos = 0; pos < nr_messages; pos++ ) {
                while( pos == w.index1_data()[i+1] )
                    i++;
                size_t j = w.index2_data()[pos];
                double w_ij = w.value_data()[pos];
                // \mu_{j\to i} = \atanh \tanh w_{ij} \tanh (\theta_j + \sum_{k\in\nb{j}\setm i} \mu_{k\to j})
                double field = sum(row(message,j)) - message(j,i) + th[j];
                double new_message = atanh( tanh( w_ij ) * tanh( field ) );
                diffs.push( fabs(message(i,j) - new_message) );
                message(i,j) = new_message;
            }

            if( verbose >= 3 )
                cout << "BinaryPairwiseGM::doBP:  maxdiff " << diffs.max() << " after " << _iterations+1 << " passes" << endl;
        }

        m = ublasvector(N);
        for( size_t j = 0; j < N; j++ ) {
            // m_j = \tanh (\theta_j + \sum_{k\in\nb{j}} \mu_{k\to j})
            double field = sum(row(message,j)) + th[j];
            m[j] = tanh( field );
        }

        if( verbose >= 1 ) {
            if( diffs.max() > tol ) {
                if( verbose == 1 )
                    cout << endl;
                    cout << "BinaryPairwiseGM::doBP:  WARNING: not converged within " << maxiter << " passes (" << toc() - tic << " clocks)...final maxdiff:" << diffs.max() << endl;
            } else {
                if( verbose >= 3 )
                    cout << "BinaryPairwiseGM::doBP:  ";
                    cout << "converged in " << _iterations << " passes (" << toc() - tic << " clocks)." << endl;
            }
        }

        return diffs.max();
    }


    double BoundJaakkolaJordanUpper( const BinaryPairwiseGM &net, ublasvector &mu ) {
        matrix W(net.w);
        ublasvector TH(net.th);
        size_t N = net.N;

        bool mu_auto = false;
        if( mu.size() == 0 ) {
            mu_auto = true;
            mu = boost::numeric::ublas::zero_vector<double>(N);
        }

        double bound = 0.0;
        for( size_t n = 0; n < N; n++ ) {
            if( mu_auto )
                mu[n] = TH[n];

            double diag = 0.0;
            for( size_t i = 0; i < N; i++ )
                diag += 0.5 * W(i,n) * W(i,n);
            double k = diag + 0.5 * pow(TH[n] - mu[n] + tanh(mu[n]),2.0) - 0.5 * pow(tanh(mu[n]),2.0) + fabs(mu[n]) + log(1.0 + exp(-2.0 * fabs(mu[n])));
            // log(2 * cosh(x)) = fabs(x) + log(1 + exp(-2 * fabs(x)));
            bound += k;

            // w_{ij}' = w_{ij} + w_{in} w_{nj}
            matrix newW(W);
            for( size_t i = 0; i < N; i++ )
                for( size_t j = 0; j < N; j++ )
                    if( i != j )
                        newW(i,j) += W(i,n) * W(n,j);
            for( size_t i = 0; i < N; i++ ) {
                newW(i,n) = 0.0;
                newW(n,i) = 0.0;
            }
            // \theta_i' = \theta_i + w_{in} (\theta_n - \mu_n + \tanh \mu_n)
            ublasvector newTH(TH);
            for( size_t i = 0; i < N; i++ )
                newTH[i] += W(i,n) * (TH[n] - mu[n] + tanh(mu[n]));
            newTH[n] = 0.0;

            W  = newW;
            TH = newTH;
        }

        return bound + net.logZ0;
    }


    // it seems that mu[n] = mf.belief(fg.var(n))[1] is an optimum for mu
    double BoundJaakkolaJordanLower( const BinaryPairwiseGM &net, ublasvector &mu ) {
        size_t N = net.N;

        // switch to 0-1 valued variables
        matrix J(net.w * 4.0);
        ublasvector h(N);
        for( size_t i = 0; i < N; i++ )
            h[i] = -2.0 * sum(row(net.w,i));
        double B = -0.25 * sum(h);
        for( size_t i = 0; i < N; i++ ) {
            h[i] += 2.0 * net.th[i];
            B -= net.th[i];
        }

        bool mu_auto = false;
        if( mu.size() == 0 ) {
            mu_auto = true;
            mu = ublasvector(N);
            for( size_t i = 0; i < N; i++ )
                mu[i] = 0.0;
        }

        double bound = B;
        for( size_t n = 0; n < N; n++ ) {
            if( mu_auto )
                mu[n] = (1.0 + tanh(h[n])) / 2.0;

            double k = mu[n] * h[n] - mu[n] * log(mu[n]) - (1.0 - mu[n]) * log(1.0 - mu[n]);
            bound += k;

            h += mu[n] * row(J,n);
        }
        return bound + net.logZ0;
    }


    // LeisinkKappen01 ("A tighter bound for graphical models", NIPS 2001)
    // equation (10) with mu_i given as input
    // more detailed equations can be found in "Graphical models and their
    // uncertainties", PhD thesis M. Leisink"
    double BoundLeisink3Lower( const BinaryPairwiseGM &net, ublasvector &mu ) {
        size_t N = net.N;

        // FIXME: use BLAS operations for below
        
        ublasvector m(N); // m_i = tanh(mu_i)
        for( size_t i = 0; i < N; i++ )
            m[i] = tanh(mu[i]);
        ublasvector nu(N); // nu_i = 1 - tanh(mu_i)^2
        for( size_t i = 0; i < N; i++ )
            nu[i] = 1.0 - m[i] * m[i];
        ublasvector alpha(N); // alpha_i = th[i] + w_{ij} tanh(mu[j]) - mu(i)
        for( size_t i = 0; i < N; i++ ) {
            alpha[i] = net.th[i] - mu[i];
            for( size_t j = 0; j < N; j++ )
                alpha[i] += net.w(i,j) * m[j];
        }

        double E2 = 0.0;
        for( size_t i = 0; i < N; i++ ) {
            E2 += alpha[i] * alpha[i] * nu[i];
            for( size_t j = 0; j < N; j++ )
                E2 += 0.5 * net.w(i,j) * net.w(i,j) * nu[i] * nu[j];
        }

        double E3 = 0.0;
        for( size_t i = 0; i < N; i++ ) {
            E3 += 2.0 * pow(alpha[i],3.0) * m[i] * nu[i];
            for( size_t j = 0; j < N; j++ ) {
                E3 -= 2.0 * pow(net.w(i,j),3.0) * m[i] * m[j] * nu[i] * nu[j];
                E3 -= 3.0 * alpha[i] * alpha[j] * net.w(i,j) * nu[i] * nu[j];
                E3 += 6.0 * alpha[i] * net.w(i,j) * net.w(i,j) * m[i] * nu[i] * nu[j];
                for( size_t k = 0; k < N; k++ )
                    E3 -= net.w(i,j) * net.w(j,k) * net.w(k,i) * nu[i] * nu[j] * nu[k];
            }
        }

        double mu_0 = 0.0;
        for( size_t i = 0; i < N; i++ ) {
            mu_0 += (net.th[i] - mu[i]) * m[i];
            for( size_t j = 0; j < N; j++ )
                mu_0 += 0.5 * net.w(i,j) * m[i] * m[j];
        }
        double lambda_0 = -E3 / (3.0 * E2);

        double logZ_mu = mu_0;
        for( size_t i = 0; i < N; i++ )
            logZ_mu += log( 2.0 * cosh(mu[i]) );

        return logZ_mu + log( 1.0 + 0.5 * exp(lambda_0) * E2 ) + net.logZ0;
    }


    double WrapBoundJaakkolaJordanUpper( const gsl_vector *v, void *params ) {
        BinaryPairwiseGM *pars = (BinaryPairwiseGM *)params;
        ublasvector mu(pars->N);
        for( size_t i = 0; i < pars->N; i++ )
            mu[i] = gsl_vector_get(v, i);
        return BoundJaakkolaJordanUpper(*pars,mu);
    }
    double WrapBoundJaakkolaJordanLower( const gsl_vector *v, void *params ) {
        BinaryPairwiseGM *pars = (BinaryPairwiseGM *)params;
        ublasvector mu(pars->N);
        for( size_t i = 0; i < pars->N; i++ )
            mu[i] = gsl_vector_get(v, i);
        return -BoundJaakkolaJordanLower(*pars,mu);
    }
    double WrapBoundLeisink3Lower( const gsl_vector *v, void *params ) {
        BinaryPairwiseGM *pars = (BinaryPairwiseGM *)params;
        ublasvector mu(pars->N);
        for( size_t i = 0; i < pars->N; i++ )
            mu[i] = gsl_vector_get(v, i);
        return -BoundLeisink3Lower(*pars,mu);
    }
    double OptimizeBound( const BinaryPairwiseGM &net, ublasvector &mu, string method ) {
        size_t N = net.N;

        Diffs diffs( N * 3, INFINITY );

        const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;
        gsl_multimin_fminimizer *s = NULL;
        gsl_vector *ss, *x0; // step size and starting point
        gsl_multimin_function minex_func;

        int status;
        double size;
        double inisize = 1.0;   // was 0.1 for JJ_UP

        // Initial vertex size vector 
        ss = gsl_vector_alloc (N);
        gsl_vector_set_all (ss, inisize);

        // Starting point
        x0 = gsl_vector_alloc (N);
        double bestf = 0.0;
        if( method == "JJ_LOW" )
            bestf = BoundJaakkolaJordanLower( net, mu );
        else if( method == "LK" )
            bestf = BoundLeisink3Lower( net, mu );
        else if( method == "JJ_UP" )
            bestf = BoundJaakkolaJordanUpper( net, mu );
        if( isnan( bestf ) )
            return bestf;

    //    cout.precision(10);
    //    cout << bestf << endl;
        for( size_t i = 0; i < N; i++ )
            gsl_vector_set (x0, i, mu[i] - 0.5 * inisize);

        // Initialize method and iterate
        if( method == "JJ_LOW" )
            minex_func.f = &WrapBoundJaakkolaJordanLower;
        else if( method == "LK" )
            minex_func.f = &WrapBoundLeisink3Lower;
        else if( method == "JJ_UP" ) 
            minex_func.f = &WrapBoundJaakkolaJordanUpper;
        minex_func.n = N;
        minex_func.params = (void *)&net;

        s = gsl_multimin_fminimizer_alloc (T, N);
        gsl_multimin_fminimizer_set (s, &minex_func, x0, ss);

        size_t iter = 0;
        double prev_diffs_max = INFINITY;
        do {
            status = gsl_multimin_fminimizer_iterate(s);

            if (status)
                break;

            size = gsl_multimin_fminimizer_size (s);
            status = gsl_multimin_test_size (size, 1e-8);

            diffs.push( size );

            iter++;
            if( iter % diffs.max_size() == 0 ) {
                if( diffs.max() == prev_diffs_max ) {
    //                cout << "seems to be stuck" << endl;
                    break;
                }
                prev_diffs_max = diffs.max();
            }
        } while (status == GSL_CONTINUE && iter < 100000 );

        double result;
        if( method == "JJ_UP" ) {
            result = s->fval;
            assert( result <= bestf );
        } else {
            result = -s->fval;
            assert( result >= bestf );
        }

        gsl_vector_free(x0);
        gsl_vector_free(ss);
        gsl_multimin_fminimizer_free (s);

        return result;
    }


    double CalcDobrushinBinary( const BinaryPairwiseGM &net, matrix &C ) {
        // Calculate C_{ij} matrix from Dobrushin's condition for the binary pairwise case
        // and return c = sup_i sum_{j\in\delta i} C_{ij}
        // See also arXiv:cond-mat/0408378v2
        
        size_t N = net.N;

        C = matrix(N, N, net.w.nnz());

        for( size_t i = 0; i < N; i++ )
            for( size_t j = 0; j < N; j++ ) {
                vector<size_t> di_j;
                vector<double> Ji;
                for( size_t k = 0; k < N; k++ )
                    if( k != i && k != j && net.w(i,k) != 0.0 ) {
                        di_j.push_back( k );
                        Ji.push_back( net.w(i,k) );
                    }
                size_t n = di_j.size();
                assert( n <= 31 );

                if( i != j && net.w(i,j) != 0.0 ) {
                    double H_ij = INFINITY;
                    for( size_t x = 0; x < (1UL << n); x++ ) {
                        double sum = net.th[i];
                        for( size_t k = 0; k < n; k++ )
                            if( x & (1UL << k))
                                sum += Ji[k];
                            else
                                sum -= Ji[k];
                        if( fabs(sum) < H_ij )
                            H_ij = fabs(sum);
                    }
                    C(i,j) = (tanh( fabs(net.w(i,j)) - H_ij ) + tanh( fabs(net.w(i,j)) + H_ij )) / 2.0;
                }
            }

        double c = 0.0;
        for( size_t i = 0; i < N; i++ ) {
            double C_i = 0.0;
            for( size_t j = 0; j < N; j++ )
                C_i += C(i,j);
            if( C_i > c )
                c = C_i;
        }

        return c;
    }


    void findLoopClustersPairwise( const FactorGraph & fg, std::set<VarSet> &allcl, VarSet newcl, const Var & root, size_t length, VarSet vars ) {
        for( VarSet::const_iterator in = vars.begin(); in != vars.end(); in++ ) {
            VarSet ind = fg.delta( *in );
            if( (newcl.size() >= 2) && ind.contains( root ) ) {
                allcl.insert( newcl + *in );
            }
            else if( length > 1 )
                findLoopClustersPairwise( fg, allcl, newcl + *in, root, length - 1, ind - newcl );
        }
    }


    size_t countLoopsPairwise( const FactorGraph & fg, size_t loopdepth ) {
        set<VarSet> loops;
        for( vector<Var>::const_iterator i0 = fg.vars().begin(); i0 != fg.vars().end(); i0++ ) {
            VarSet i0d = fg.delta(*i0);
            if( loopdepth > 1 )
                findLoopClustersPairwise( fg, loops, *i0, *i0, loopdepth - 1, i0d );
        }
        return loops.size();
    }


    double CalcGirthPairwise( const FactorGraph &fg ) {
        bool girth_reached = false;
        size_t loopdepth;
        for( loopdepth = 2; loopdepth <= fg.nrVars() && !girth_reached; loopdepth++ ) {
            size_t nr_loops = countLoopsPairwise( fg, loopdepth );
            if( nr_loops > 0 )
                girth_reached = true;
        }
        if( girth_reached )
            return loopdepth - 1;
        else 
            return INFINITY;
    }


    // Proposition V.6 in Tatikonda, "Convergence of the Sum-Product Algorithm", 2003
    double BoundTatikonda( double c, double M ) {
        if( M == INFINITY )
            return 0.0;
        else if( c >= 1.0 )
            return 1.0;
        else {
            double bound = (1.0 + M / 2.0) * pow((1.0 + 2.0 / M) * c, M / 2.0);
            return( (bound < 1.0) ? bound : 1.0 );
        }
    }

    
    // Bounds distance from BP marginal
    //   |p(x_i) - b_i(x_i)| \le min{1,c_i}  where  c_i = \sum_j C_{ij}
    // according to Taga and Mase, "Error Bounds Between Marginal 
    // Probabilities and Beliefs of Loopy Belief Propagataion Algorithm",
    // MICAI 2007, LNAI 4293, pp. 186-106, 2006, Springer-Verlag
    //
    // CalcDobrushinBinary( net, C ) should be called first.
    double BoundTagaMase( const matrix &C, size_t i ) {
        double c_i = sum(row(C,i));
        return( (1.0 < c_i) ? 1.0 : c_i );
    }


/*
    // Bounds distance from BP marginal
    //   |p(x_i) - b_i(x_i)| \le min{1,c_i}  where  c_i = \sum_j C_{ij}
    // according to Taga and Mase, "Error Bounds Between Marginal 
    // Probabilities and Beliefs of Loopy Belief Propagataion Algorithm",
    // MICAI 2007, LNAI 4293, pp. 186-106, 2006, Springer-Verlag
    //
    // CalcDobrushinBinary( net, C ) should be called first.
    double BoundTagaMaseImproved( const FactorGraph &fg, const matrix &C, size_t i ) {
//        BinaryPairwiseGM net( fg );
//        matrix C_Dobrushin;
//        double c = CalcDobrushinBinary( net, C_Dobrushin );

        Bounds fb(fg, Properties()("tol",string("1e-9"))("verbose",string("0"))("useSAW",string("0"))("useIhler",string("0")) );
        fb.init();
        fb.buildGraph( fg.var(i) );
        VarSet Sprime = fb.VarsWithOriginalNeighborhood();
//        cerr << "i = " << fg.var(i) << ", Sprime = " << Sprime << endl;

        size_t N = fg.nrVars();
        ublasvector a(N);
        ublasvector new_a(N);
        for( size_t j = 0; j < N; j++ )
            a(j) = 1.0;
        
        for( size_t iter = 0; iter < 10; iter++ ) {
            ublasvector Ca(N);
            Ca = prod(C, a);
            for( size_t j = 0; j < N; j++ ) {
                if( Sprime >> fg.var(j) )
                    new_a(j) = Ca(j);
                else
                    new_a(j) = Ca(j) + 1.0;
                
                if( new_a(j) < a(j) )
                    a(j) = new_a(j);
            }
        }
        
        return a(i);
    }
*/


    // Calculate a lower or upper bound on logZ
    double BoundlogZ( const FactorGraph &fg, string method, bool debug_bounds, size_t verbose) {
        double outertol = 1e-6;
        Properties opts = Properties()("maxiter",(size_t)10000)("tol",1e-9)("verbose",verbose);
        Properties mf_opts = opts;
        Properties bp_opts = opts("updates",string("SEQMAX"));
        Properties jt_opts = opts("updates", string("HUGIN"));
        Properties trw_opts = opts("maxiter",(size_t)100000)("updates",string("SEQFIX"))("outertol",outertol)("maxouteriter",(size_t)100);
//        trw_opts.Set("tol",1e-7);  I used this instead of 1e-9 for the case of very strong interactions in test_bounds_new
            
        if( method == "BP" ) {
            BP bp(fg, bp_opts);
            bp.init();
            bp.run();
            return real(bp.logZ());
        } else if( method == "EXACT" ) {
            JTree jt( fg, jt_opts );
            jt.init();
            jt.run();
            return real(jt.logZ());
        } else if( method == "TRW" ) {
            TRW trw( fg, trw_opts );
            trw.init();
            double result = trw.run();
            if( result > outertol ) {
                cerr << "TRW not converged" << endl;
                return INFINITY;
            }
            else
                return real(trw.logZ());
        } else if( method == "MF" ) {
            MF mf( fg, mf_opts );
            mf.init();
            mf.run();
            return real(mf.logZ());
        } else if( method == "JJ" || method == "JJOPT" ) {
            BinaryPairwiseGM net( fg );
            ublasvector mu;
            if( method == "JJ" ) {
                double b = BoundJaakkolaJordanUpper( net, mu );
                if( isnan( b ) )
                    return INFINITY;
                else
                    return b;
            } else {
                double b = OptimizeBound( net, mu, "JJ_UP" );
                if( isnan( b ) )
                    return INFINITY;
                else
                    return b;
            }
        } else if( method == "LK" || method == "LKOPT" ) {
            MF mf( fg, mf_opts );
            mf.init();
            mf.run();
            BinaryPairwiseGM net( fg );
            ublasvector mu( fg.nrVars() );
            for( size_t i = 0; i < fg.nrVars(); i++ )
                mu[i] = 0.5 * log(mf.belief(fg.var(i))[1] / mf.belief(fg.var(i))[0]);
            if( method == "LK" ) 
                return BoundLeisink3Lower( net, mu );
            else
                return OptimizeBound( net, mu, "LK" );
        } else if( method == "LCLO" || method == "LCUP" ) {
            BinaryPairwiseGM net( fg );
            double w_max = 0.0;
            size_t d = 0;
            size_t nrEdges = 0;
            for( size_t i = 0; i < net.N; i++ ) {
                size_t d_i = 0;
                for( size_t j = 0; j < net.N; j++ ) {
                    if( net.w(i,j) > 0.0 ) {
                        d_i++;
                        if( i < j )
                            nrEdges++;
                    }
                    if( net.w(i,j) > w_max )
                        w_max = net.w(i,j);
                }
                if( d_i > d )
                    d = d_i;
            }
            double psi_max = exp(2.0 * w_max);
            double psi_bar = std::pow(2.0, (double)d) * std::pow(std::pow(psi_max,3) - 1, (double)d) * std::pow(psi_max, 2.0*d*d);

            BP bp(fg, bp_opts);
            bp.init();
            bp.run();
            double logZBethe = real(bp.logZ());

            // This is not an exact bound, a term of order O(log(nrEdges) / nrEdges) is still missing

            if( method == "LCLO" ) 
                return logZBethe - nrEdges * log(1.0 + psi_bar); // + nrEdges * O(log(nrEdges)/nrEdges)
            else
                return logZBethe + nrEdges * log(1.0 + psi_bar); // + nrEdges * O(log(nrEdges)/nrEdges)
        } else
            DAI_THROW(UNKNOWN_DAI_ALGORITHM);
    }


    // Tree-Reparameterization Bounds of Wainwright et al
    void BoundTRP( const FactorGraph &fgin, vector<Box> &results ) {
        assert( fgin.isPairwise() );
        FactorGraph fg = fgin.maximalFactors();

        Properties opts = Properties()("maxiter",(size_t)10000)("tol",1e-9)("verbose",(size_t)0);


        // start with running BP
        Properties bp_opts = opts("updates",string("SEQFIX"));
        BP bp(fg, bp_opts);
        bp.init();
        bp.run();

        // build the TRP version of the original factor graph
        // and build the weighted Markov graph, where weights on edges
        // correspond to the strength of the interaction on TRP
        vector<Factor> trp_fac;
        trp_fac.reserve( fg.nrFactors() + fg.nrVars() );
        for( size_t i = 0; i < fg.nrVars(); i++ )
            trp_fac.push_back( bp.beliefV( i ) );
        WeightedGraph<double> wg;
        double maxWeight = 0.0;
        for( size_t I = 0; I < fg.nrFactors(); I++ ) {
            VarSet::const_iterator i = fg.factor(I).vars().begin();
            size_t i1 = fg.findVar( *i );
            size_t i2 = fg.findVar( *(++i) );
            Factor pairfac = (bp.beliefF( I ) / (bp.beliefV( i1 ) * bp.beliefV( i2 )));
            trp_fac.push_back( pairfac );

            double w = pairfac.strength( fg.var( i1 ), fg.var( i2 ) );
            if( w > maxWeight )
                maxWeight = w;
            wg[UEdge(i1, i2)] = -w;
        }
        FactorGraph trp( trp_fac );
        // add constant because BGL implementation of Prim's algorithm cannot handle negative weights
        for( WeightedGraph<double>::iterator e = wg.begin(); e != wg.end(); e++ )
            e->second += maxWeight;
        // find maximal spanning tree
        DEdgeVec mst = MinSpanningTreePrims( wg );

        // find removed edges and build factor graph corresponding to mst
        set<size_t> removed_edges;
        vector<Factor> tree_fac;
        tree_fac.reserve( mst.size() + fg.nrVars() );
        for( size_t i = 0; i < fg.nrVars(); i++ )
            tree_fac.push_back( bp.beliefV( i ) );
        for( size_t I = 0; I < fg.nrFactors(); I++ ) {
            bool removed = true;
            VarSet::const_iterator i = fg.factor(I).vars().begin();
            size_t i1 = fg.findVar( *i );
            size_t i2 = fg.findVar( *(++i) );
            for( DEdgeVec::const_iterator e = mst.begin(); e != mst.end(); e++ )
                if( UEdge( *e ) == UEdge( i1, i2 ) ) {
                    removed = false;
                    break;
                }
            if( removed )
                removed_edges.insert( I );
            else
                tree_fac.push_back( trp.factor( I + fg.nrVars() ) );
        }

        // create factor graph by projecting trp on the mst
        FactorGraph trp_tree( tree_fac );
        BP trp_tree_bp(trp_tree, bp_opts);
        trp_tree_bp.init();
        trp_tree_bp.run();
        // calculate some marginals on the tree
        map<VarSet,Factor> trp_tree_marginals;
        for( size_t _s = 0; _s < fg.nrVars(); _s++ )
            trp_tree_marginals[fg.var(_s)] = trp_tree_bp.beliefV(_s);
        for( set<size_t>::const_iterator _I = removed_edges.begin(); _I != removed_edges.end(); _I++ ) {
            size_t I = *_I;
            VarSet Ivars = fg.factor(I).vars();
            trp_tree_marginals[Ivars] = calcMarginal( trp_tree_bp, Ivars, false );
            for( size_t _s = 0; _s < fg.nrVars(); _s++ ) {
                Var s = fg.var(_s);
                trp_tree_marginals[Ivars + s] = calcMarginal( trp_tree_bp, Ivars + s, false );
            }
        }

        // calculate Deltas
        vector<Factor> Delta;
        Delta.reserve( fg.nrVars() );
        for( size_t _s = 0; _s < fg.nrVars(); _s++ ) {
            Var s = fg.var(_s);
            Delta.push_back( Factor( s, 0.0 ) );
            for( set<size_t>::const_iterator _I = removed_edges.begin(); _I != removed_edges.end(); _I++ ) {
                size_t I = *_I;
                VarSet Ivars = fg.factor(I).vars();
                Factor cov = trp_tree_marginals[Ivars + s] - trp_tree_marginals[Ivars - s] * trp_tree_marginals[s];
                Delta[_s] += (trp.factor(I).log() * cov).notSum(s);
            }
        }

        // calculate Phi(theta*)
        double lZ = 0.0;
        // calculate upper bound using TRW
        double outertol = 1e-6;
        Properties trw_opts = opts("maxiter",(size_t)100000)("updates",string("SEQFIX"))("outertol",outertol)("maxouteriter",(size_t)100);
        TRW trw( trp, trw_opts );
        trw.init();
        double result = trw.run();
        if( result > outertol ) {
            cerr << "TRW not converged" << endl;
            return;
        }
        lZ = real(trw.logZ());

        // calculate upper bound on D( trp_tree | bp)
        double D = lZ;
        for( set<size_t>::const_iterator _I = removed_edges.begin(); _I != removed_edges.end(); _I++ ) {
            size_t I = *_I;
            VarSet Ivars = fg.factor(I).vars();
            D -= (trp_tree_marginals[Ivars] * trp.factor(I).log()).totalSum();
        }

        // calculate for each variable a lower and upper bound
        results = vector<Box>();
        results.reserve( fg.nrVars() );
        for( size_t i = 0; i < fg.nrVars(); i++ ) {
            Factor id( fg.var(i), 1.0 );
            Factor bi = bp.beliefV( i );
            Factor lb = bi * (Delta[i] / bi - D).exp();
            Factor ub = id - (id - bi) * ((Delta[i] / (id - bi) + D) * -1.0).exp();
            Box b( fg.var(i), lb.p(), ub.p() );
            results.push_back( b );
        }
    }


}
