/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <fstream>
#include "alldai.h"
#include "util.h"


using namespace std;
using namespace libDAI;


map<size_t, size_t> ReadEvidence( char *filename ) {
    map<size_t, size_t> evid;

    ifstream is;
    is.open( filename );
    if( is.is_open() ) {
        try { // FIXME: no exceptions are generated by iostreams by default
            size_t nr_evid;
            is >> nr_evid;

            for( size_t i = 0; i < nr_evid; i++ ) {
                size_t label, val;
                is >> label;
                is >> val;
                evid[label] = val;
            }

            is.close();
        } catch (...) {
            is.close();
            throw;
        }
    } else
        DAI_THROW(CANNOT_READ_FILE);

    return evid;
}


pair<vector<Var>, vector<Factor> > ReadFromUAIFile( const char *filename, long verbose ) {
    vector<Factor> factors;
    vector<Var> vars;

    ifstream is;
    is.open( filename );
    if( is.is_open() ) {
        try { // FIXME: no exceptions are generated by iostreams by default
            size_t nr_f, nr_v;
            string line;
            
            getline(is,line);
//            if( line != "BAYES" && line != "MARKOV" )
//                DAI_THROW(INVALID_FACTORGRAPH_FILE);
            if( verbose >= 2 )
                cout << "Reading " << line << " network..." << endl;

            is >> nr_v;
            if( is.fail() )
                DAI_THROW(INVALID_FACTORGRAPH_FILE);
            if( verbose >= 2 )
                cout << "Reading " << nr_v << " variables..." << endl;

            for( size_t i = 0; i < nr_v; i++ ) {
                size_t dim;
                is >> dim;
                if( is.fail() )
                    DAI_THROW(INVALID_FACTORGRAPH_FILE);
                vars.push_back( Var( i, dim ) );
            }

            is >> nr_f;
            if( is.fail() )
                DAI_THROW(INVALID_FACTORGRAPH_FILE);
            if( verbose >= 2 )
                cout << "Reading " << nr_f << " factors..." << endl;

            vector<vector<long> > labels;
            for( size_t I = 0; I < nr_f; I++ ) {
                if( verbose >= 3 )
                    cout << "Reading factor " << I << "..." << endl;
                size_t I_nrvars;
                is >> I_nrvars;
                if( verbose >= 3 )
                    cout << "  I_nrvars: " << I_nrvars << endl;

                vector<long> I_labels;
                vector<size_t> I_dims;
                VarSet I_vars;
                for( size_t _i = 0; _i < I_nrvars; _i++ ) {
                    long label;
                    is >> label;
                    I_labels.push_back(label);
                    I_dims.push_back(vars[label].states());
                    I_vars += vars[label];
                }
                if( verbose >= 3 )
                    cout << "  labels: " << I_labels << ", dimensions " << I_dims << endl;

                // add the Factor and the labels
                factors.push_back(Factor(I_vars,0.0));
                labels.push_back(I_labels);
                
            }

            for( size_t I = 0; I < nr_f; I++ ) {
                if( verbose >= 3 )
                    cout << "Reading factor " << I << "..." << endl;

                reverse(labels[I].begin(), labels[I].end());    // last label is least significant

                size_t I_nrvars = factors[I].vars().size();
                vector<size_t> I_dims;
                for( size_t _i = 0; _i < I_nrvars; _i++ )
                    I_dims.push_back(vars[labels[I][_i]].states());
                if( verbose >= 3 )
                    cout << "  labels: " << labels[I] << ", dimensions " << I_dims << endl;

                // calculate permutation sigma (internally, members are sorted)
                vector<size_t> sigma(I_nrvars,0);
                VarSet::const_iterator j = factors[I].vars().begin();
                for( size_t mi = 0; mi < I_nrvars; mi++,j++ ) {
                    long search_for = j->label();
                    vector<long>::iterator j_loc = find(labels[I].begin(),labels[I].end(),search_for);
                    sigma[mi] = j_loc - labels[I].begin();
                }
                if( verbose >= 3 )
                    cout << "  sigma: " << sigma << endl;

                // calculate multindices
                Permute permindex( I_dims, sigma );
                
                // read values
                size_t nr_nonzeros;
                is >> nr_nonzeros;
                if( verbose >= 3 ) 
                    cout << "  nonzeroes: " << nr_nonzeros << endl;
                assert( nr_nonzeros == factors[I].states() );
                for( size_t li = 0; li < nr_nonzeros; li++ ) {
                    double val;
                    is >> val;
                    factors[I][permindex.convert_linear_index( li )] = val;
                }
            }

            if( verbose >= 3 )
                cout << "factors:" << factors << endl;


            is.close();
        } catch (...) {
            is.close();
            throw;
        }
    } else
        DAI_THROW(CANNOT_READ_FILE);

    return pair<vector<Var>, vector<Factor> >(vars,factors);
}


int main( int argc, char *argv[] ) {
    if ( argc != 4 ) {
        cout << "Usage: " << argv[0] << " <filename.uai> <filename.uai.evid> <seed>" << endl << endl;
        cout << "Reads factor graph <filename.uai> and evidence <filename.uai.evid>" << endl;
        cout << "Solver for UAI approximate inference competition" << endl;
        return 1;
    } else {
        rnd_seed( atoi(argv[3]) );
        long verbose = 0;


        pair<vector<Var>, vector<Factor> > input;
        input = ReadFromUAIFile( argv[1], verbose );
        map<size_t,size_t> evid = ReadEvidence( argv[2] );
        double logZcorr = 0.0;
        if( 1 ) {
            for( size_t I = 0; I < input.second.size(); I++ ) {
                for( map<size_t,size_t>::const_iterator e = evid.begin(); e != evid.end(); e++ ) {
                    if( input.second[I].vars() >> input.first[e->first] ) {
                        if( verbose >= 2 )
                            cout << "clamping " << e->first << " to value " << e->second << " in factor " << I << " = " << input.second[I].vars() << endl;
                        input.second[I] = input.second[I].slice( input.first[e->first], e->second );
                        if( verbose >= 2 )
                            cout << "...remaining vars: " << input.second[I].vars() << endl;
                    }
                }
            }
            // remove empty factors
            for( vector<Factor>::iterator I = input.second.begin(); I != input.second.end(); )
                if( I->vars().size() == 0 ) {
                    logZcorr += log( (*I)[0] );
                    I = input.second.erase( I );
                } else
                    I++;
        } else {
            for( map<size_t,size_t>::const_iterator e = evid.begin(); e != evid.end(); e++ ) {
                // add delta factors
                input.second.push_back( Factor( input.first[e->first], 0.0 ) );
                input.second.back()[e->second] = 1.0;
            }
        }
        FactorGraph fg(input.second);


        Properties opts;
        opts.Set("maxiter",10000U);
        opts.Set("tol",1e-9);
        opts.Set("verbose",0U);

        InferenceAlgorithm *alg;
        double tic = toc();

#ifdef SOLVER1
        JTree jt( fg, opts("updates",string("HUGIN")) );
        alg = &jt;
#endif
#ifdef SOLVER2
        BP bp(fg, opts("updates",string("SEQMAX")));
        alg = &bp;
#endif
#ifdef SOLVER3
    for( size_t ijgp_i = 1; ijgp_i < 100; ijgp_i++ ) {
        IJGP ijgp(fg, opts("updates",string("SEQFIX"))("i",ijgp_i));
        alg = &ijgp;
#endif
#ifdef SOLVER4
        MF mf(fg, opts);
        alg = &mf;
#endif
#ifdef SOLVER5
        TreeEP te(fg, opts("type",string("ORG")));
        alg = &te;
#endif
#ifdef SOLVER6
        HAK hak(fg, opts("doubleloop",string("1"))("clusters",string("MIN")));
        alg = &hak;
#endif
#ifdef SOLVER7
        LC lcbp(fg, opts("cavity",string("PAIR"))("reinit",string("0"))("updates",string("SEQFIX"))("cavainame",string("BP"))("cavaiopts",string("[updates=SEQMAX,tol=1e-9,maxiter=10000,verbose=0]")));
        alg = &lcbp;
#endif
#ifdef SOLVER8
        LCLin lcbplin(fg, opts("cavity",string("PAIR"))("reinit",string("0"))("updates",string("SEQFIX"))("cavainame",string("BP"))("cavaiopts",string("[updates=SEQMAX,tol=1e-9,maxiter=10000,verbose=0]")));
        alg = &lcbplin;
#endif
#ifdef SOLVER9
        LC lcbp(fg, opts("cavity",string("FULL"))("reinit",string("0"))("updates",string("SEQFIX"))("cavainame",string("BP"))("cavaiopts",string("[updates=SEQMAX,tol=1e-9,maxiter=10000,verbose=0]")));
        alg = &lcbp;
#endif

        alg->init();
        alg->run();

#ifdef UAIPR
        cout.precision( 8 );
        cout << "z " << (real( alg->logZ() ) + logZcorr) / log(10) << endl;
#endif
#ifdef UAIMAR
        cout << "m " << input.first.size();
        for( size_t i = 0; i < input.first.size(); i++ ) {
            cout << " " << input.first[i].states();
            if( evid.find( i ) != evid.end() ) {
                for( size_t state = 0; state < input.first[i].states(); state++ ) 
                    if( state == evid[i] )
                        cout << " 1";
                    else
                        cout << " 0";
            } else {
                Factor b = alg->belief(input.first[i]);
                for( size_t state = 0; state < input.first[i].states(); state++ )
                    cout << " " << b[state];
            }
        }
        cout << endl;
#endif
        cout << "t " << toc() - tic << endl;

#ifdef SOLVER3
    }
#endif
    }

    return 0;
}
