/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include "factorgraph.h"
#include "jtree.h"
#include "bp.h"
#include "daialg.h"
#include <iostream>
#include <stdlib.h>


using namespace std;
using namespace libDAI;


int main( int argc, char *argv[] ) {
    if( argc != 4 ) {
        cout << "Usage: " << argv[0] << " <in.fg> <cavityvarind> <onlycavity>" << endl << endl;
        cout << "Converts a .fg (FactorGraph) file to a .dot (GraphViz) file for visualization," << endl;
        cout << "making a cavity graph corresponding to the variable with index <cavityvarind>." << endl;
        cout << "If <onlycavity>!=0 then the rest of the factor graph is not drawn." << endl;
        cout << "The .dot file can be converted to .ps (PostScript) by 'neato -T ps out.dot > out.ps'" << endl;
        return 1;
    } else {
        // Read factorgraph
        FactorGraph fg;
        char *infile = argv[1];
        fg.ReadFromFile( infile );

        size_t cavityvarind = atoi( argv[2] );
        size_t onlycavity = atoi( argv[3] );
        Var cavityvar = fg.var(cavityvarind);
        VarSet delta = fg.delta(cavityvar);

//            JTree jt( fg, Properties()("updates",string("HUGIN"))("verbose", 0UL) );

        for( cavityvarind = 0; cavityvarind < fg.nrVars(); cavityvarind++ ) {
            Var cavityvar = fg.var(cavityvarind);
            VarSet delta = fg.delta(cavityvar);
            BP jt( fg, Properties()("updates",string("SEQMAX"))("verbose", 0UL)("tol", 1e-9)("maxiter", 1000UL));
            jt.grm().makeCavity( cavityvar );
            vector<Factor> pairbeliefs;
            pairbeliefs = calcPairBeliefs( jt, jt.grm().delta(cavityvar), true );
            for( size_t ind = 0; ind < pairbeliefs.size(); ind++ ) {
                Var i = *(pairbeliefs[ind].vars().begin());
                Var j = *(pairbeliefs[ind].vars() - i).begin();
                pairbeliefs[ind].normalize();
                double strength = real( KL_dist( pairbeliefs[ind], pairbeliefs[ind].marginal(i) * pairbeliefs[ind].marginal(j)) );
                if( strength > 0.02 )
                    cout << fg.var(cavityvarind) << ": " << i.label() << "--" << j.label() << ": " << strength << endl;
            }
        }
    
        /*
        cout << "graph G {" << endl;
        cout << "graph[size=\"10,10\"];" << endl;

        if( !onlycavity ) {
            cout << "node[shape=circle,width=0.4,fixedsize=true];" << endl;
            for( size_t i = 0; i < fg.nrVars(); i++ )
                if( !((delta | cavityvar) >> fg.var(i)) )
                    cout << "\tx" << fg.var(i).label() << ";" << endl;
        }
        cout << "node[shape=circle,width=0.4,fixedsize=true,color=purple];" << endl;
        cout << "\tx" << cavityvar.label() << ";" << endl;
        cout << "node[shape=circle,width=0.4,fixedsize=true,color=red];" << endl;
        for( size_t i = 0; i < fg.nrVars(); i++ )
            if( delta >> fg.var(i) )
                cout << "\tx" << fg.var(i).label() << ";" << endl;

        if( !onlycavity ) {
            cout << "node[shape=box,style=filled,color=lightgrey,width=0.3,height=0.3,fixedsize=true];" << endl;
            for( size_t I = 0; I < fg.nrFactors(); I++ ) 
                if( !(fg.factor(I).vars() >> cavityvar) )
                    cout << "\tp" << I << ";" << endl;
        }
        cout << "node[shape=box,style=filled,color=purple,width=0.3,height=0.3,fixedsize=true];" << endl;
        for( size_t I = 0; I < fg.nrFactors(); I++ ) 
            if( (fg.factor(I).vars() >> cavityvar) )
                cout << "\tp" << I << ";" << endl;

        if( !onlycavity ) {
            cout << "edge[color=blue];" << endl;
            for( size_t iI = 0; iI < fg.nrEdges(); iI++ )
                if( !(fg.factor(fg.edge(iI).second).vars() >> cavityvar) )
                    cout << "\tx" << fg.var(fg.edge(iI).first).label() << " -- p" << fg.edge(iI).second << ";" << endl;
        }
        cout << "edge[color=purple];" << endl;
        for( size_t iI = 0; iI < fg.nrEdges(); iI++ )
            if( (fg.factor(fg.edge(iI).second).vars() >> cavityvar) )
                cout << "\tx" << fg.var(fg.edge(iI).first).label() << " -- p" << fg.edge(iI).second << ";" << endl;

        for( size_t ind = 0; ind < pairbeliefs.size(); ind++ ) {
            Var i = *(pairbeliefs[ind].vars().begin());
            Var j = *(pairbeliefs[ind].vars() / i).begin();
            pairbeliefs[ind].normalize();
            double strength = real( KL_dist( pairbeliefs[ind], pairbeliefs[ind].marginal(i) * pairbeliefs[ind].marginal(j)) );
            if( strength < 0.0 )
                strength = 0.0;
            cout << "edge[weight=10,color=gray" << (100-int(strength * 100)) << "] x" << i.label() << " -- x" << j.label() << ";" << endl;
        }

        cout << "}" << endl;

        */
        return 0;
    }
}
