#include "pindex.hh"
#include "stat.hh"

#include "fgraph.hh"

#include <iostream>
#include <vector>
#include <algorithm>
#include <fstream>
#include <iomanip>
#include "testai.hh"
#include "jtree.h"

#include <boost/dynamic_bitset.hpp>

typedef	boost::dynamic_bitset<unsigned long> Bitset;
typedef std::vector<unsigned char> Degrees;

using namespace std;

int main(int argc, char* argv[])
{
	int ret = -1;

  try {
    string filename;
    string aliases;
    vector<string> methods;
    double tol;
    size_t all_comb = 0;
    size_t maxiter;
    size_t verbose;
 
    int prec = 15, R = 0;

    po::options_description opts_required("Required options", 100);
    opts_required.add_options()
      ("filename", po::value< string >(&filename), "Filename of FactorGraph")
      ("methods", po::value< vector<string> >(&methods)->multitoken(), "AI methods to test")
      ("all_comb", po::value< size_t >(&all_comb), "if 0 computes Z_2 corrections. If 1, all combinations of triplets (default 0)")
      ;

    po::options_description opts_optional("Allowed options", 100);
    opts_optional.add_options()
      ("help", "produce help message")
      ("aliases", po::value< string >(&aliases), "Filename for aliases")
      ("tol", po::value< double >(&tol), "Override tolerance")
      ("maxiter", po::value< size_t >(&maxiter), "Override maximum number of iterations")
      ("verbose", po::value< size_t >(&verbose), "Override verbosity")
      ("R", po::value< int >(&R), "random seed. If R!=0, BP messages are initialized randomly")
      ;

    po::options_description cmdline_options(100);
    cmdline_options.add(opts_required).add(opts_optional);

    po::variables_map vm;
    po::store(po::parse_command_line(argc, argv, cmdline_options), vm);      	    
    po::notify(vm);

    if( vm.count("help") || !(vm.count("filename") && vm.count("methods")) ) {
      cout << "Reads factorgraph <filename.fg> and performs the approximate" << endl;
      cout << "inference algorithms <method*>, reporting clocks, max and average" << endl;
      cout << "error and relative logZ error (comparing with the results of" << endl;
      cout << "<method0>, the base method, for which one can use JTREE_HUGIN)." << endl << endl;
      cout << opts_required << opts_optional << endl;
      return 1;
    }

    Real bptol = tol;
    uint bpmaxiter = 1000;
    int variant = 2, tries = 0;

    int max_clique_size = -1;
		double cpu_jtree = 0;

    // Create loop graph
    FGraph graph(filename, (TBPVariant)variant, prec, R, verbose);

		// Run BP (try the following schedulings)
    bool conv = false;
    while (!conv && (tries < 5) ) {

			if (verbose) {
        switch (variant) {
          case 1: cerr << "BP_SEQFIXED" << endl; break;  
          case 2: cerr << "BP_SEQRANDOM" << endl; break;  
          case 3: cerr << "BP_SEQMAXRES" << endl; break;  
          case 4: cerr << "BP_PAR" << endl; break;  
        }
			}
      graph.setTBPVariant( (TBPVariant) variant );
      if (graph.doBP(bptol, bpmaxiter, false)) {
        conv = true;
			}
      else {
        if (verbose) cerr << graph.m_maxdif << endl;
        variant = (variant+1)%4;
        tries++;
      }
    }

    if (conv) {

			// compute the Pfaffian Series in the Forney representation
			// toForneyStyle already sets the result of BP in the Forney graph so we don't need to run
			// BP in the Forney graph which takes more iterations (and sometimes worse convergence
			graph.toForneyStyle();
			graph.doPfaffianSeries( all_comb );

			if (Abs(graph.getLogZBPForney() - graph.getLogZBP()) > Real(1e-10)) {
				cerr << "LogZ differs between Forney (" << graph.getLogZBPForney() << ") and Factor (" << graph.getLogZBP();
				cerr << ") diff = " << Abs(graph.getLogZBPForney() - graph.getLogZBP()) << endl;
			}

      // Read aliases
      map<string,string> Aliases;
      if( !aliases.empty() ) {
        ifstream infile;
        infile.open (aliases.c_str());
        if (infile.is_open()) {
          while( true ) {
            string line;
            getline(infile,line);
            if( infile.fail() )
              break;
            if( (!line.empty()) && (line[0] != '#') ) {
              string::size_type pos = line.find(':',0);
              if( pos == string::npos )
                throw "Invalid alias";
              else {
                string::size_type posl = line.substr(0, pos).find_last_not_of(" \t");
                string key = line.substr(0, posl + 1);
                string::size_type posr = line.substr(pos + 1, line.length()).find_first_not_of(" \t");
                string val = line.substr(pos + 1 + posr, line.length());
                Aliases[key] = val;
              }
            }
          }
          infile.close();
        } else
          throw "Error opening aliases file";
      } 

      FactorGraph fg;
      if( fg.ReadFromFile(filename.c_str()) ) {
        cerr << "Error reading " << filename << endl;
        return 2;
      } else {         
				pair<int,int> tw = treewidth(fg);
				max_clique_size = tw.first;			


        vector<Factor> q0;
        double logZ0 = 0.0;

        for( size_t m = 0; m < methods.size(); m++ ) {
          pair<string, Properties> meth = parseMethod( methods[m], Aliases );

					if( vm.count("tol") )
            meth.second.Set("tol",tol);
          if( vm.count("maxiter") )
            meth.second.Set("maxiter",maxiter);
          if( vm.count("verbose") )
            meth.second.Set("verbose",verbose);
          TestDAI piet(fg, meth.first, meth.second );

					double init_clock; 
          if( meth.first == string("JTREE") ) init_clock = toc();
          piet.doDAI();
          if( meth.first == string("JTREE") ) cpu_jtree = ( toc() - init_clock );

					// Joris considers that the first method is the exact
					if ( m == 0 ) {        
            q0 = piet.q;
						logZ0 = piet.logZ;
						if( meth.first == string("JTREE") ) {
							
							// I consider the JTREE as the only exact method
							// set the exact results into the loopgraph object
							//max_clique_size = ((JTree *)piet.getInfAlg())->getMaxClique();
							
							for (uint i=0; i<q0.size(); i++) {
								VProb pi(Real(2.));
								//for (uint v=0; v<q0[i].stateSpace(); v++) {
								for (uint v=0; v<q0[i].states(); v++) {
									pi[v] = Real(q0[i][v]);
								}
								graph.setExactMarginal(i, pi);
							}
							graph.m_logZ_exact = logZ0;
//cout << "EXACT Z = " << setprecision(15) << logZ0 << " " << exp(logZ0) << endl;
//exit(-1);
						}
					}

					piet.calcErrs(q0);
        }
      }

			// output results cout
			graph.outputRes();

			// output results cerr
			Real abs_logZ = Abs(graph.m_logZ_exact);
			ret = ((Abs(graph.getLogZPfaffian() - abs_logZ) / abs_logZ) < Real(1e-14)) ? 0 : 1;

			cout << setprecision(14) << scientific;

			cout << graph.getBPVariant() << " ";	// 1 
			/////////////
			cout << graph.getCpuBP() << " ";	// 2
			cout << graph.getLogZBP() << " "; // 4
			/////////////
			cout << max_clique_size << " "; // 5
			cout << cpu_jtree << " "; // 6
			cout << graph.getLogZExact() << " "; // 8
			/////////////
			cout << graph.getCpuPfaffian1() << " "; // 9
			cout << graph.getLogZPfaffian1() << " "; // 11
			/////////////
			cout << graph.getCpuPfaffian() << " "; // 12
			cout << graph.getLogZPfaffian() << " "; // 14
			/////////////
			cout << "0 "; // 15
			cout << "0 "; // 16

		} // end convergence 
		else {

			cout << graph.getBPVariant() << " "; 
			/////////////
			cout << " -1 ";
			cout << " -1 ";
			/////////////
			cout << max_clique_size << " ";
			cout << cpu_jtree << " ";
			cout << graph.getLogZExact() << " ";
			/////////////
			cout << " -1 ";
			cout << " -1 ";
			/////////////
			cout << " -1 ";
			cout << " -1 ";
			/////////////
			cout << " -1 ";
			cout << " -1\t";

			return ret;
		}
  } catch(const char *e) {
    cerr << "Exception: " << e << endl;
    return 1;
  } catch(std::exception& e) {
    cerr << "Exception: " << e.what() << endl;
    return 1;
  }
  catch(...) { 
    cerr << "Exception of unknown type!" << endl;
  }

  return ret;
}


//				cout << graph.getSLoopSize() << " ";
//				cout << graph.getGLoopSize() << " ";
//				cout << graph.getNoLoopSize() << " ";
//				cout << graph.getMergings() << " ";
//				cout << graph.getCpuTLSBP() << " ";
//				cout << toSeconds( graph.getCpuTLSBP() ) << " ";
//				cout << graph.m_nloops_ok << " ";
//				cout << graph.m_nloops << " ";

